#!/usr/bin/env python3
import unittest
from subprocess import PIPE, run
import os
import shutil
import sys
import argparse


# This is not best practice but for testing, this is the best I could
# come up with
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# TODO: Species specific aqquired genes only pheno results, not spec specific?

test_names = ["test1", "test2", "test3", "test4"]
test_data = {
    # Test published acquired resistance
    test_names[0]: "data/test_isolate_01.fa",
    test_names[1]: "data/test_isolate_01_1.fq data/test_isolate_01_2.fq",
    # Test published point mut resistance
    test_names[2]: "data/test_isolate_05.fa",
    test_names[3]: "data/test_isolate_05_1.fq data/test_isolate_05_2.fq",
}
run_test_dir = "running_test"
working_dir = os.path.dirname(os.path.realpath(__file__))


class ResFinderRunTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Delete "running_test" folder from previous tests if still exists
        if os.path.isdir(run_test_dir):
            try:
                shutil.rmtree(run_test_dir)
            # The following error has occured using VirtualBox under Windows 10
            # with ResFinder installed in a shared folder:
            #   OSError [Errno: 26] Text file busy: 'tmp'
            except OSError:
                procs = run(["rm", "-r", run_test_dir])

        # Set absolute path for database folders and external programs
        cls.db_path_res = os.path.abspath(args.db_path_res)
        cls.blastPath = os.path.abspath(args.blast_path)
        cls.kmaPath = os.path.abspath(args.kma_path)
        cls.db_path_point = os.path.abspath(args.db_path_point)
        cls.dir_res = os.path.join(os.path.dirname(__file__), '../', )
        cls.dir_res = os.path.abspath(cls.dir_res)
        # Change working dir to test dir
        os.chdir(working_dir)
        # Does not allow running two tests in parallel
        os.makedirs(run_test_dir, exist_ok=False)

    @classmethod
    def tearDownClass(cls):
        try:
            shutil.rmtree(run_test_dir)
        # The following error has occured using VirtualBox under Windows 10
        # with ResFinder installed in a shared folder:
        #   OSError [Errno: 26] Text file busy: 'tmp'
        except OSError:
            procs = run(["rm", "-r", run_test_dir])

    def test_on_data_with_just_acquired_resgene_using_blast(self):
        # Maria has an E. coli isolate, with unknown resistance.
        # At first, she just wants to know which acquired resistance genes are
        # found in the genome.
        # She therefore runs resfinder cmd line.

        # First Maria checks out the documentation.
        procs = run("python3 ../run_resfinder.py -h", shell=True, stdout=PIPE,
                    check=True)
        output = procs.stdout.decode()
        self.assertIn("--help", output)

        # Maria goes on to run ResFinder for acquired genes with her E. coli
        # isolate.
        # First she creates a few directories to store her output.
        test1_dir = run_test_dir + "/" + test_names[0]
        os.makedirs(test1_dir)
        # Then she runs run_resfinder with her first isolate.
        cmd_acquired = ("python3  " + self.dir_res + "/run_resfinder.py"
                        + " -ifa " + test_data[test_names[0]]
                        + " -o " + test1_dir
                        + " -s 'Escherichia coli'"
                        + " --min_cov 0.6"
                        + " -t 0.8"
                        + " --acquired"
                        + " --db_path_res " + self.db_path_res
                        + " --blastPath " + self.blastPath)

        procs = run(cmd_acquired, shell=True, stdout=PIPE, stderr=PIPE,
                    check=True)

        fsa_hit = test1_dir + "/ResFinder_Hit_in_genome_seq.fsa"
        fsa_res = test1_dir + "/ResFinder_Resistance_gene_seq.fsa"
        res_table = test1_dir + "/ResFinder_results_table.txt"
        res_tab = test1_dir + "/ResFinder_results_tab.txt"
        results = test1_dir + "/ResFinder_results.txt"

        with open(fsa_hit, "r") as fh:
            check_result = fh.readline()
        self.assertIn("blaB-2_1_AF189300", check_result)

        with open(fsa_res, "r") as fh:
            check_result = fh.readline()
        self.assertIn("blaB-2_AF189300", check_result)

        with open(res_table, "r") as fh:
            for line in fh:
                if(line.startswith("blaB-2")):
                    check_result = line
                    break
        self.assertIn("blaB-2_1_AF189300", check_result)

        with open(res_tab, "r") as fh:
            fh.readline()
            check_result = fh.readline()
        self.assertIn("blaB-2_1_AF189300", check_result)

        with open(results, "r") as fh:
            fh.readline()
            fh.readline()
            fh.readline()
            fh.readline()
            fh.readline()
            check_result = fh.readline()
        self.assertIn("blaB-2_1_AF189300", check_result)

    def test_on_data_with_just_acquired_resgene_using_kma(self):
        # Maria has another E. coli isolate, with unknown resistance.
        # This time she does not have an assembly, but only raw data.
        # She therefore runs resfinder cmd line using KMA.

        # First she creates a few directories to store her output.
        test2_dir = run_test_dir + "/" + test_names[1]
        os.makedirs(test2_dir, exist_ok=False)

        # Then she runs run_resfinder with her first isolate.
        cmd_acquired = ("python3 " + self.dir_res + "/run_resfinder.py"
                        + " -ifq " + test_data[test_names[1]]
                        + " -o " + test2_dir
                        + " -s 'Escherichia coli'"
                        + " --min_cov 0.6"
                        + " -t 0.8"
                        + " --acquired"
                        + " --db_path_res " + self.db_path_res
                        + " --kmaPath " + self.kmaPath)

        procs = run(cmd_acquired, shell=True, stdout=PIPE, stderr=PIPE,
                    check=True)

        fsa_hit = test2_dir + "/ResFinder_Hit_in_genome_seq.fsa"
        fsa_res = test2_dir + "/ResFinder_Resistance_gene_seq.fsa"
        res_table = test2_dir + "/ResFinder_results_table.txt"
        res_tab = test2_dir + "/ResFinder_results_tab.txt"
        results = test2_dir + "/ResFinder_results.txt"

        with open(fsa_hit, "r") as fh:
            check_result = fh.readline()
        self.assertIn("blaB-2", check_result)

        with open(fsa_res, "r") as fh:
            check_result = fh.readline()
        self.assertIn("blaB-2_AF189300", check_result)

        with open(res_table, "r") as fh:
            for line in fh:
                if(line.startswith("blaB-2")):
                    check_result = line
                    break
        self.assertIn("blaB-2", check_result)

        with open(res_tab, "r") as fh:
            fh.readline()
            check_result = fh.readline()
        self.assertIn("blaB-2", check_result)

        with open(results, "r") as fh:
            fh.readline()
            fh.readline()
            fh.readline()
            fh.readline()
            fh.readline()
            check_result = fh.readline()
        self.assertIn("blaB-2", check_result)

    def test_on_data_with_just_point_mut_using_blast(self):
        # Maria also wants to check her assembled E. coli isolate for
        # resistance caused by point mutations.

        # First she creates a few directories to store her output.
        test3_dir = run_test_dir + "/" + test_names[2]
        os.makedirs(test3_dir)

        # Then she runs run_resfinder with her first isolate.
        cmd_point = ("python3 " + self.dir_res + "/run_resfinder.py"
                     + " -ifa " + test_data[test_names[2]]
                     + " -o " + test3_dir
                     + " -s 'Escherichia coli'"
                     + " --min_cov 0.6"
                     + " --threshold 0.8"
                     + " --point"
                     + " --db_path_point " + self.db_path_point
                     + " --db_path_res " + self.db_path_res
                     + " --blastPath " + self.blastPath)

        procs = run(cmd_point, shell=True, stdout=PIPE, stderr=PIPE,
                    check=True)

        # Expected output files
        pf_pred = test3_dir + "/PointFinder_prediction.txt"
        pf_res = test3_dir + "/PointFinder_results.txt"
        pf_table = test3_dir + "/PointFinder_table.txt"

        with open(pf_res, "r") as fh:
            fh.readline()
            check_result = fh.readline()
        self.assertIn("gyrA", check_result)
        self.assertIn("p.S83A", check_result)

        point_mut_found = False
        with open(pf_table, "r") as fh:
            for line in fh:
                if(line.startswith("gyrA p.S83A")):
                    check_result = line
                    point_mut_found = True
                    break
        self.assertEqual(point_mut_found, True)

    def test_on_data_with_just_point_mut_using_kma(self):
        # Maria has another E. coli isolate, with unknown resistance.
        # This time she does not have an assembly, but only raw data.
        # She therefore runs resfinder cmd line using KMA.

        # First she creates a few directories to store her output.
        test4_dir = run_test_dir + "/" + test_names[3]
        os.makedirs(test4_dir, exist_ok=False)

        # Then she runs run_resfinder with her first isolate.
        cmd_acquired = ("python3 " + self.dir_res + "/run_resfinder.py"
                        + " -ifq " + test_data[test_names[3]]
                        + " -o " + test4_dir
                        + " -s 'Escherichia coli'"
                        + " --min_cov 0.6"
                        + " --threshold 0.8"
                        + " --point"
                        + " --db_path_point " + self.db_path_point
                        + " --db_path_res " + self.db_path_res
                        + " --kmaPath " + self.kmaPath)

        procs = run(cmd_acquired, shell=True, stdout=PIPE, stderr=PIPE,
                    check=True)

        # Expected output files
        pf_pred = test4_dir + "/PointFinder_prediction.txt"
        pf_res = test4_dir + "/PointFinder_results.txt"
        pf_table = test4_dir + "/PointFinder_table.txt"

        with open(pf_res, "r") as fh:
            fh.readline()
            check_result = fh.readline()
        self.assertIn("gyrA", check_result)
        self.assertIn("p.S83A", check_result)

        point_mut_found = False
        with open(pf_table, "r") as fh:
            for line in fh:
                if(line.startswith("gyrA p.S83A")):
                    check_result = line
                    point_mut_found = True
                    break
        self.assertEqual(point_mut_found, True)


def parse_args():
    parser = argparse.ArgumentParser(add_help=False, allow_abbrev=False)
    group = parser.add_argument_group("Options")
    group.add_argument('-res_help', "--resfinder_help",
                       action="help")
    group.add_argument("-db_res", "--db_path_res",
                       help="Path to the databases for ResFinder",
                       default="./db_resfinder")
    group.add_argument("-b", "--blastPath",
                       dest="blast_path",
                       help="Path to blastn",
                       default="./cge/blastn")
    group.add_argument("-k", "--kmaPath",
                       dest="kma_path",
                       help="Path to KMA",
                       default="./cge/kma/kma")
    group.add_argument("-db_point", "--db_path_point",
                       help="Path to the databases for PointFinder",
                       default="./db_pointfinder")
    ns, args = parser.parse_known_args(namespace=unittest)
    return ns, sys.argv[:1] + args

if __name__ == "__main__":
    args, argv = parse_args()   # run this first
    sys.argv[:] = argv       # create cleans argv for main()
    unittest.main()



#!/usr/bin/env python3

import argparse
import os.path
import re
import shutil
from signal import *
import tempfile
import sys
import subprocess
from itertools import chain

from .feature import Feature, ResGene, Mutation, ResMutation
from .phenodbpoint import PhenoDBPoint
from .res_profile import PhenoDB, ResProfile, FeatureGroup
from .dbhit import DBHit
from ..output.exceptions import DatabaseError


class Isolate(dict):
    """ An isolate class is a dict of Features.
    """
    NO_AB_CLASS = "No class defined"

    def __init__(self, name, species=None, amr_panel_file=None):
        self.name = name
        self.resprofile = None
        self.feature_classes = {}
        self.species = None
        if(species is not None):
            self.amr_panel = Isolate.load_amr_panel(species, amr_panel_file)
        else:
            self.amr_panel = None

    @staticmethod
    def load_amr_panel(species, panel_file):
        """
        """
        panels = {}
        inclusions = {}

        with open(panel_file, "r") as fh:
            panel_name = None
            for line in fh:

                line = line.rstrip()

                # Skip empty lines and comments
                if not line or line.startswith("#"):
                    continue

                # Get panel name
                new_panel_name = Isolate._get_panel_name(panels, line)
                if new_panel_name is not None:
                    panel_name = new_panel_name
                    continue

                if Isolate._get_inclusions(panel_name, line, inclusions):
                    continue

                # Get Antibiotics
                if(panel_name):
                    Isolate._get_antibiotics(line, panel_name, panels)

        Isolate._merge_inclusions(panels, inclusions)

        species_key = Isolate.check_panel_name(species, panels)

        return set(panels[species_key]) if species_key else None

    @staticmethod
    def _get_antibiotics(line, panel_name, panels):
        """ Stores list of antimicrobials in dict of panel names. """
        tmp_list = panels.get(panel_name, [])
        tmp_list.append(line.lower())
        panels[panel_name] = tmp_list

    @staticmethod
    def _get_panel_name(panels, line):
        match_panel = re.search(r':Panel:\s{0,1}(.+)$', line)
        if(match_panel):
            panel_name = match_panel.group(1).lower()
            panels[panel_name] = []
            return panel_name
        else:
            return None

    @staticmethod
    def _get_inclusions(panel_name, line, inclusions):
        match_inclusion = re.search(r':Include:\s{0,1}(.+)$', line)
        if(match_inclusion):
            include_panel = match_inclusion.group(1).lower()
            tmp_list = inclusions.get(panel_name, [])
            tmp_list.append(include_panel)
            inclusions[panel_name] = tmp_list
            return True
        else:
            return False

    @staticmethod
    def _merge_inclusions(panels, inclusions):
        for panel, include_list in inclusions.items():
            panel_list = panels[panel]
            include_abs = []
            for incl_panel in include_list:
                include_abs = include_abs + panels[incl_panel]
            panels[panel] = panel_list + include_abs

    @staticmethod
    def check_panel_name(name, panels):
        """ Panel names are expected to consist of "Genus species" or
            only "Genus". The method checks the name against the loaded
            panel names and returns the panel name that matches the
            given name. If no panel name matches it checks if the first
            word of the given name matches any of the genus panel
            names, and considers that a match if found.

            Returns False if no match is found
            Returns None if no panels has been loaded
            Returns the panel name that match
        """
        if(not panels):
            return None

        name = name.lower()
        if(name in panels):
            return name

        genus_name = " ".split(name)[0]
        if(genus_name in panels):
            return genus_name

        return False

    def load_resfinder_tab(self, tabbed_output, phenodb):
        with open(tabbed_output, "r") as fh:
            while(True):
                line = fh.readline()
                if(not line):
                    break

                line = line.rstrip()
                if(not line):
                    continue

                db_name = line
                second_line = fh.readline().rstrip()

                if(second_line == "No hit found"):
                    continue

                # At this point second line must be headers, and are skipped.

                res_hit = fh.readline().rstrip()

                while(res_hit):
                    hit_list = res_hit.split("\t")
                    match_length, ref_length = hit_list[2].split("/")
                    start_ref, end_ref = hit_list[4].split("..")
                    accno = hit_list[8]
                    gene_name = hit_list[0]
                    unique_id = "{0}_{1}".format(gene_name, accno)
                    hit = DBHit(name=gene_name, identity=hit_list[1],
                                match_length=match_length,
                                ref_length=ref_length, start_ref=start_ref,
                                end_ref=end_ref, acc=accno,
                                db="resfinder")

                    start_feat, end_feat = hit_list[6].split("..")

                    if(start_feat == "NA"):
                        start_feat = None
                        end_feat = None

                    phenotypes, unique_id = self.get_phenotypes(phenodb,
                                                                unique_id)
                    ab_class = set()
                    if(phenotypes):
                        for p in phenotypes:
                            for ab in p.antibiotics:
                                ab_class.update(ab.classes)
                    else:
                        ab_class.add(db_name)

                    gene_feat = ResGene(unique_id=unique_id,
                                        seq_region=hit_list[5],
                                        start=start_feat, end=end_feat,
                                        hit=hit, ab_class=ab_class)

                    if(unique_id in self):
                        temp_list = self[unique_id]
                        temp_list.append(gene_feat)
                        self[unique_id] = temp_list
                    else:
                        self[unique_id] = [gene_feat]

                    res_hit = fh.readline().rstrip()

    @staticmethod
    def get_phenodb_id(feat_res_dict, type):
        """
            Input:
                feat_res_dict: entry in the dict 'genes' or 'seq_variations'
                               from a cgecore Result object.
                type: Which dict in the Result object, either 'seq_regions' or
                      'seq_variations'
            Output: Key formatted as the ones in the PhenoDB object.

            Method is used to reformat an id from a Result object, so that it
            can be used to query a PhenoDB object.

            TODO: Make this method obsolete by harmonizing the ids in PhenoDB
                  with the ones from the Result object.
        """
        if(type == "seq_variations"):
            var_aa = feat_res_dict.get("var_aa", None)
            var_codon = feat_res_dict.get("var_codon", None)
            codon_change = feat_res_dict.get("codon_change", None)
            indels = feat_res_dict['deletion'] or feat_res_dict['insertion']

            # Not point mutation
            if(var_aa is None and var_codon is None):
                return feat_res_dict["seq_regions"][0]
            # RNA mutation(single nucleotide)
            elif(len(var_codon)==1 and codon_change is None):
                nuc_format = (f"{feat_res_dict['seq_regions'][0]}"
                              f"_{feat_res_dict['ref_start_pos']}"
                              f"_{feat_res_dict['var_codon']}")
                aa_format = ""
                return nuc_format, aa_format
            # indels
            elif (indels):
                nuc_format = (f"{feat_res_dict['seq_regions'][0]}"
                              f"_{feat_res_dict['ref_end_pos']}"
                              f"_{feat_res_dict['nuc_change']}")
                aa_format = (f"{feat_res_dict['seq_regions'][0]}"
                             f"_{feat_res_dict['ref_start_pos']}"
                             f"_{var_aa}")  # var_aa can be None
                return nuc_format, aa_format
            # Amino acid mutation
            else:
                nuc_format = ""
                aa_format = (f"{feat_res_dict['seq_regions'][0]}"
                             f"_{feat_res_dict['ref_start_pos']}"
                             f"_{feat_res_dict['var_aa']}")
                return nuc_format, aa_format

        elif(type == "seq_regions"):
            return ("{}_{}".format(feat_res_dict["name"],
                                  feat_res_dict["ref_acc"]), "")

    def load_finder_results(self, std_table, phenodb, type):
        """
            Input:
                std_table: Result object from cgecore module loaded with
                           ResFinder and/or Pointfinder results.
                phenodb: PhenoDB object
                type: 'seq_regions' or 'seq_variations', for ResFinder and
                      PointFinder results, respectively.

            Method loads Isolate object
        """
        for key, feat_info in std_table[type].items():
            # Skip genes from PointFinder database (not resistance genes).
            if(type == "seq_regions"
               and any(re.search("PointFinder", entry)
                       for entry in feat_info["ref_database"])):
                continue

            unique_id_nuc, unique_id_aa = Isolate.get_phenodb_id(feat_info,
                                                                 type)
            phenotypes, unique_id = self.get_phenotypes(phenodb, unique_id_nuc,
                                                        unique_id_aa, type)
            feat_list = self.get(unique_id, [])
            if(phenotypes):
                for p in phenotypes:
                    res_feature = self.new_res_feature(type, feat_info,
                                                       unique_id,
                                                       unique_id_aa,
                                                       unique_id_nuc, p)
                    if(res_feature not in feat_list):
                        feat_list.append(res_feature)
            else:
                res_feature = self.new_res_feature(type, feat_info,
                                                   unique_id,
                                                   unique_id_aa,
                                                   unique_id_nuc)
                feat_list.append(res_feature)
            if unique_id_nuc == "":
                self[unique_id_aa] = feat_list
            else:
                self[unique_id_nuc] = feat_list

    def get_phenotypes(self, phenodb, unique_id_nuc, unique_id_aa, type):
        """
            Input:
                phenodb: PhenoDB object
                unique_id: string in key format fitting the phenodb objects
            Output:
                returns a phenodb object based on the unique id and the unique
                id used.
            Method modifies unique id to account for mutation type (AA/NUC)
            using the pointfinder database.
        """
        if (phenodb.mut_type_is_defined
                and type == "seq_variations"):
            unique_id_aa = unique_id_aa + '_AA'
            unique_id_nuc = unique_id_nuc + '_NUC'

        phenotype_aa = phenodb.get(unique_id_aa, None)
        phenotype_nuc = phenodb.get(unique_id_nuc, None)

        if phenotype_nuc:
            phenotypes = phenotype_nuc
            unique_id_found = unique_id_nuc
        else:
            phenotypes = phenotype_aa
            unique_id_found = unique_id_aa

        return phenotypes, unique_id_found

    def new_res_feature(self, type, feat_info, unique_id, aa_format, nuc_format,
                        phenotype=None):
        ab_class = set()

        if phenotype is None:
            ab_class.add(Isolate.NO_AB_CLASS)
            db_pheno = None
            if type == "seq_regions":
                ref_id = feat_info.get("ref_id", None)
                raise DatabaseError(
                    f"Phenotype not found for sequence region: {ref_id}.\n"
                    "If this application was run with an official database, "
                    "please contact the curators.\n"
                    "If this application was run with a custom database, "
                    "please check that your phenotype file contains all "
                    "sequence regions.")
        else:
            for ab in phenotype.antibiotics:
                ab_class.update(ab.classes)
            db_pheno = phenotype.res_database

        if(type == "seq_regions"):
            res_feature = self.new_res_gene(feat_info, unique_id, ab_class,
                                            phenotype)
        elif(type == "seq_variations"):
            res_feature = self.new_res_mut(feat_info, unique_id, ab_class,
                                           phenotype, nuc_format, aa_format)

        ResProfile.update_classes_dict_of_feature_sets(
            self.feature_classes, res_feature)
        return res_feature

    def new_res_mut(self, feat_info, unique_id, ab_class, phenotype, nuc_format,
                    aa_format):
        ref_aa = feat_info.get("ref_aa", None)

        if(ref_aa is None or ref_aa.upper() == "NA"):
            nucleotide_mut = True
        else:
            nucleotide_mut = False

        if phenotype:
            feat_res = ResMutation(
                unique_id=unique_id,
                nuc_format=nuc_format,
                aa_format=aa_format,
                seq_region=";;".join(feat_info["seq_regions"]),
                pos=feat_info["ref_start_pos"],
                ref_codon=feat_info["ref_codon"],
                mut_codon=feat_info["var_codon"],
                ref_aa=feat_info.get("ref_aa", None),
                mut_aa=feat_info.get("var_aa", None),
                isolate=self,
                insertion=feat_info["insertion"],
                deletion=feat_info["deletion"],
                end=feat_info["ref_end_pos"],
                nuc=nucleotide_mut,
                ab_class=ab_class,
                pmids=phenotype.pmid,
                notes=phenotype.notes,
                ref_db=phenotype.res_database
            )
        else:
            feat_res = ResMutation(
                unique_id=unique_id,
                nuc_format= nuc_format,
                seq_region=";;".join(feat_info["seq_regions"]),
                pos=feat_info["ref_start_pos"],
                ref_codon=feat_info["ref_codon"],
                mut_codon=feat_info["var_codon"],
                ref_aa=feat_info.get("ref_aa", None),
                mut_aa=feat_info.get("var_aa", None),
                isolate=self,
                insertion=feat_info["insertion"],
                deletion=feat_info["deletion"],
                end=feat_info["ref_end_pos"],
                nuc=nucleotide_mut,
                ab_class=ab_class
            )

        return feat_res

    def new_res_gene(self, gene_info, unique_id, ab_class, phenotype):
        hit = DBHit(name=gene_info["name"],
                    identity=gene_info["identity"],
                    match_length=gene_info["alignment_length"],
                    ref_length=gene_info["ref_seq_length"],
                    start_ref=gene_info["ref_start_pos"],
                    end_ref=gene_info["ref_end_pos"],
                    acc=gene_info["ref_acc"],
                    depth=gene_info.get("depth", None),
                    db=phenotype.res_database)

        query_start = gene_info.get("query_start_pos", None)
        query_end = gene_info.get("query_end_pos", None)
        query_id = gene_info.get("query_id", None)

        feat_res = ResGene(unique_id=unique_id,
                           seq_region=query_id,
                           start=query_start,
                           end=query_end,
                           hit=hit,
                           ab_class=ab_class,
                           pmids=phenotype.pmid,
                           notes=phenotype.notes,
                           ref_db=phenotype.res_database)
        return feat_res

    def calc_res_profile(self, phenodb):
        """
        """
        features = self.values()
        # Flatten features.
        features = list(chain.from_iterable(features))
        self.resprofile = ResProfile(features, phenodb)

    def profile_to_str_table(self, header=False):
        """
        """
        output_str = ""

        if(header):
            output_str = (
                "# ResFinder phenotype results.\n"
                "# Sample: " + self.name + "\n"
                "# \n"
                "# The phenotype 'No resistance' should be interpreted with\n"
                "# caution, as it only means that nothing in the used\n"
                "# database indicate resistance, but resistance could exist\n"
                "# from 'unknown' or not yet implemented sources.\n"
                "# \n"
                "# The 'Match' column stores one of the integers 0, 1, 2, 3.\n"
                "#      0: No match found\n"
                "#      1: Match < 100% ID AND match length < ref length\n"
                "#      2: Match = 100% ID AND match length < ref length\n"
                "#      3: Match = 100% ID AND match length = ref length\n"
                "# If several hits causing the same resistance are found,\n"
                "# the highest number will be stored in the 'Match' column.\n"
                "\n"
            )
            output_str += ("# Antimicrobial\t"
                           "Class\t"
                           "WGS-predicted phenotype\t"
                           "Match\t"
                           "Genetic background\n")

        # For each antibiotic class
        for ab_class in self.resprofile.phenodb.antibiotics.keys():
            # For each antibiotic in current class
            for ab_db in self.resprofile.phenodb.antibiotics[ab_class]:
                output_str += ("{ab:s}\t{cl:s}"
                               .format(ab=ab_db.name, cl=ab_class))

                # Isolate is resistant towards the antibiotic
                if(ab_db in self.resprofile.resistance):
                    ab = self.resprofile.resistance[ab_db]
                    output_str += "\tResistant"

                    # Find the resistance causing gene with the best match
                    # Mutations will always have best match as they are only
                    # either present or absent.
                    best_match = 0
                    for unique_id in ab.features:
                        feature_list = ab.features[unique_id]
                        for feature in feature_list:
                            if(isinstance(feature, Mutation)
                               or isinstance(feature, FeatureGroup)):
                                best_match = 3
                            # Note: Mutations do not have "hits"
                            elif(feature.hit.match_category > best_match):
                                best_match = feature.hit.match_category

                    output_str += "\t" + str(best_match)

                    gene_list = ab.get_gene_namewacc(tostring=True)
                    mut_list = ab.get_mut_namewannot(tostring=True)

                    if(gene_list and mut_list):
                        gene_mut_str = gene_list + " " + mut_list
                    elif(gene_list):
                        gene_mut_str = gene_list
                    elif(mut_list):
                        gene_mut_str = mut_list
                    else:
                        gene_mut_str = ""

                    output_str += "\t" + gene_mut_str + "\n"

                # TODO: delete elif clause.
                # Isolate is susceptibile towards the antibiotic
                elif(ab_db.name in self.resprofile.susceptibile):
                    ab = self.resprofile.susceptibile[ab_db.name]
                    # Genetic background is not written if susceptibile
                    gene_list = ""
                    # Uncomment next line to write genetic background for susc.
                    # gene_list = ab.get_gene_namewacc(tostring=True)
                    output_str += "\tNo resistance\t0\t" + gene_list + "\n"
                else:
                    output_str += "\tNo resistance\t0\t\n"

        if(self.resprofile.missing_db_features):
            output_str += ("\n# WARNING: Missing features from phenotype "
                           "database:\n")
            output_str += "# Feature_ID\tRegion\tDatabase\tHit\n"

            for feature in self.resprofile.missing_db_features:
                output_str += ("{}\t{}\t"
                               .format(feature.unique_id, feature.seq_region))

                if(feature.hit is None):
                    output_str += "\t"
                else:
                    output_str += str(feature.hit.db) + "\t" + feature.hit.name

                output_str += "\n"

        if(self.resprofile.unknown_db_features):
            output_str += ("\n# WARNING: Features with unknown phenotype \n")
            output_str += "# Feature_ID\tRegion\tDatabase\tHit\tClass\n"

            for feature in self.resprofile.unknown_db_features:
                output_str += ("{}\t{}\t"
                               .format(feature.unique_id, feature.seq_region))

                if(feature.hit is None):
                    output_str += "\t"
                else:
                    output_str += str(feature.hit.db) + "\t" + feature.hit.name

                output_str += "\t" + ",".join(feature.ab_class)

                output_str += "\n"

        return output_str



#!/usr/bin/env python3

from __future__ import print_function
import argparse
import os.path
import re
import shutil
from signal import *
import tempfile
import sys
import subprocess

from .phenotype import Phenotype


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class PhenoDBPoint(dict):
    """ Loads a text table into dict.
        The dict consists of Phenotype objects. The keys are unique ids.
    """
    def __init__(self, txt_file, alt_name_file=None):

        # Stores non-redundant complete list of antibiotics in DB.
        self.antibiotics = {}

        with open(txt_file, "r") as fh:
            # Skip headers
            fh.readline()
            line_counter = 1

            for line in fh:
                try:
                    line_counter += 1

                    line_list = line.split("\t")
                    line_list = map(str.rstrip, line_list)

                    # ID in DB is just Gene ID and is not unique
                    phenodb_id = line_list[0]
                    codon_pos = line_list[3]
                    res_codon_str = line_list[6].lower()
                    unique_id = (phenodb_id + "_" + codon_pos + "_"
                                 + res_codon_str)

                    ab_class = self.get_csv_tuple(line_list[7].lower())

                    pub_phenotype = self.get_csv_tuple(line_list[8].lower())
                    if("unknown" in pub_phenotype or "none" in pub_phenotype):
                        pub_phenotype = ()

                    pmid = self.get_csv_tuple(line_list[9].lower())

                    phenotype = list(pub_phenotype)

                    if(len(line_list) > 10 and line_list[10]):
                        sug_phenotype = self.get_csv_tuple(line_list[10])
                        for p in sug_phenotype:
                            if p not in pub_phenotype:
                                phenotype.append(p)
                    else:
                        sug_phenotype = None

                    if(len(line_list) > 11 and line_list[11]):
                        res_mechanics = line_list[11]
                    else:
                        res_mechanics = None

                    if(len(line_list) > 12 and line_list[12]):
                        susceptibile = self.get_csv_tuple(line_list[12])
                    else:
                        susceptibile = ()

                    if(len(line_list) > 13 and line_list[13]):
                        notes = line_list[13]
                    else:
                        notes = ""

                    # Create non-redundant complete list of antibiotics in DB.
                    for ab in phenotype:
                        for _class in ab_class:
                            if(_class in self.antibiotics):
                                self.antibiotics[_class][ab] = True
                            else:
                                self.antibiotics[_class] = {}
                                self.antibiotics[_class][ab] = True
                    for ab in susceptibile:
                        for _class in ab_class:
                            if(_class in self.antibiotics):
                                self.antibiotics[_class][ab] = True
                            else:
                                self.antibiotics[_class] = {}
                                self.antibiotics[_class][ab] = True

                    pheno = Phenotype(unique_id, phenotype, ab_class,
                                      sug_phenotype, pub_phenotype, pmid,
                                      susceptibile=susceptibile, notes=notes,
                                      res_mechanics=res_mechanics)

                    self[unique_id] = pheno

                    # A pointmutation with several different res codons will
                    # never be found using all the res_codons. Instead it will
                    # be found with just one.
                    # Alternative unique ids are therefore made using just a
                    # single res_codon.
                    res_codon = self.get_csv_tuple(line_list[6].lower())
                    if(len(res_codon) > 1):
                        for codon in res_codon:
                            unique_id_alt = (phenodb_id + "_" + codon_pos
                                             + "_" + codon)
                            self[unique_id_alt] = pheno
                except IndexError:
                    eprint("Error in line " + str(line_counter))
                    eprint("Split line:\n" + str(line_list))

    @staticmethod
    def get_csv_tuple(csv_string):
        """ Takes a string containing a comma seperated list, makes it all
            lower case, remove empty entries, remove trailing and preseeding
            whitespaces, and returns the list as a tuple.
        """
        csv_string = csv_string.lower()
        string_list = csv_string.split(",")
        # Remove empty entries.
        out_list = [var.strip() for var in string_list if var]
        return tuple(out_list)

    def print_db_stats(self):
        """ Prints some stats about the database to stdout.
        """
        counter_ab_class = 0
        counter_ab = 0

        ab_output = ""

        print("-------------- LIST OF CLASSES --------------")
        for ab_class in self.antibiotics:
            counter_ab_class += 1
            print(ab_class)
        print("--------------- END OF CLASSES --------------")
        print("|")
        print("|")
        print("------------ LIST OF ANTIBIOTICS ------------")
        for ab_class in self.antibiotics:
            print(ab_class)
            for ab in self.antibiotics[ab_class]:
                counter_ab += 1
                print("\t" + ab)
        print("------------ END OF ANTIBIOTICS -------------")
        print("|")
        print("|")
        print("------------------ SUMMARY ------------------")
        print("No. of classes: " + str(counter_ab_class))
        print("No. of antibiotics: " + str(counter_ab))
        print("-------------------- END --------------------")



#!/usr/bin/env python3
from .orderedset import OrderedSet
from .exceptions import DuplicateKeyError
from .exceptions import LockedObjectError


class TableResults(dict):
    """
        The hit table is created empty.
    """
    def __init__(self, software, version, run_date, run_cmd, id):
        self.software = software
        self.version = version
        self.run_date = run_date
        self.run_cmd = run_cmd
        self.id = id

        self.long = {}
        self.medium = {}
        self.short = {}
        self.databases = {}

    def add_database(self, id, version):
        self.databases[id] = version

    def add_table(self, table_name, copy=False):
        """
            TODO: Implement copy argument
        """
        if(isinstance(table_name, Table)):
            name = table_name.name
        else:
            name = table_name

        if(self.long.get(name, False)):
            raise DuplicateKeyError("All table names need to be unique. An "
                                    "attempt was made to add a table with a "
                                    "name that already exists. The name is:{}."
                                    .format(name))

        if(isinstance(table_name, Table)):
            if(copy):
                raise NotImplementedError()
                # self.long[name] = copy.deepcopy(table_name)
            else:
                self.long[name] = table_name
        else:
            self.long[name] = Table(name)

    def as_txt(self, type, header=True, sep="\t", sort=True, pre_name=""):
        if(type == "long"):
            tr = self.long
        elif(type == "medium"):
            tr = self.medium
        elif(type == "short"):
            tr = self.short
        else:
            raise ValueError("as_text method in TableResults class must be "
                             "called with either 'long', 'medium', or "
                             "'short'. It was called with {}".format(type))

        out_str = ""

        for table_name in tr:
            table_str = tr[table_name].as_txt(header=header, sep=sep, sort=sort)
            out_str += ("{pre}{name}\n{table}\n"
                        .format(pre=pre_name, name=table_name, table=table_str))

        return out_str


    def merge(self, tr2, auto_row_header=False, copy=False):
        """
            Merges another table result object into the calling table results
            instance.

            Table objects with identical names will be merged, as will headers
            within merged tables.

            The other table is copied with copy.deepcopy so that changes to the
            merged tables or changes to the other table does not affect each
            other.

            IMPORTANT: Currently ONLY merges the tables stored in 'self.long'
            IMPORTANT: Table merge function does not copy the table that gets
                       merged. Thus, a change in the table that gets merged
                       will also occur in the merged table.
            TODO: Implement the copy argument to create a deepcopy of the
                  merged table instead of just referencing it.
        """
        if(copy is True):
            raise NotImplementedError()

        for table_name in tr2.long:
            other_tbl = tr2.long[table_name]

            if(table_name not in self.long):
                self.add_table(table_name)
                self.long[table_name] = other_tbl
                continue

            call_tbl = self.long[table_name]

            for row_header in other_tbl:
                # Get row now as row header might change if a row header
                # collision occurs.
                other_row_dict = other_tbl[row_header]

                # Handle row header collisions
                if(row_header in call_tbl):
                    if(auto_row_header is False):
                        raise DuplicateKeyError(
                            "During merge of two table, row headers were "
                            "found not to be unique across the two tables. "
                            "Row headers needs to be renamed or the merge "
                            "method should be called with "
                            "auto_row_header=True. Note auto row headers will "
                            "become incrementing integers.\n"
                            "Calling table: {}\n"
                            "Other table: {}\n"
                            "Row header: {}\n"
                            .format(call_tbl.name, other_tbl.name, row_header))
                    call_tbl._row_header_count += 1
                    row_header = call_tbl._row_header_count

                call_tbl[row_header] = other_row_dict


class Row(dict):
    """
    """
    def __init__(self, table):
        self.table = table
        super(Row, self).__init__()

    def __setitem__(self, col_header, col_val):
        if(not col_header in self.table._headers):
            if(self.table.lock_headers):
                raise LockedObjectError("Attempt to add a header via a Row "
                                        "insertion to a table, where the "
                                        "headers has been locked.\n"
                                        "Table: {}\n"
                                        "Row header, val: {}, {}\n"
                                        "Header: {}\n"
                                        .format(self.name, col_header, col_val,
                                                header))
            self.table.add_header(col_header)

        super(Row, self).__setitem__(col_header, col_val)


class Table(dict):
    """
        Keys of a Table object is considered 'row headers'.
        The Key can be None in which case the row header will become an
        increasing integer (starting at 1).

        Values must be lists of length 2. Where the first entry is the name of
        a column header and the second entry is the value.
    """
    NULL_VAL = "NA"

    def __init__(self, name):
        self.name = name
        self.lock_headers = False

        self._headers = OrderedSet()
        self._row_header_count = 0
        self._sort_key = "row_header"
        self._sort_list = SortList(container=self)
        self._iter_count = 0

    def add_header(self, header, exist_ok=True):
        if(self.lock_headers):
            raise LockedObjectError("Attempt to add a header to an object "
                                    "where the headers has been locked.\n"
                                    "Table: {}\n"
                                    "Header: {}\n".format(self.name, header))

        if(exist_ok is False and header in self._headers):
            raise DuplicateKeyError("Attempt to add header that already exists"
                                    " to table, but the add_header method was "
                                    "called with exist_ok=False.\n"
                                    "Table: {}\n"
                                    "Header: {}\n".format(self.name, header))
        self._headers.add(header)

    def add_headers(self, headers, exist_ok=True):
        for header in headers:
            self.add_header(header, exist_ok)

    def as_txt(self, header=True, sep="\t", sort=True):
        """
        """
        if(sort):
            self.set_sort_key(self._sort_key)

        if(header):
            out_str = ("{sep}{heads}\n"
                       .format(sep=sep, heads=sep.join(self._headers)))
        else:
            out_str = ""

        for row in self:
            row_list = self.get_row_as_list(row, as_txt=True)
            out_str += ("{}\n".format(sep.join(row_list)))

        return out_str


    def extract_column(self, col):
        """
        """
        if(col == "row_header"):
            return list(self.keys())
        if(col not in self._headers):
            raise KeyError("Attempt to extract a column that doesn't exist.\n"
                           "Missing column name: {}\n"
                           "Table name: {}".format(col, self.name))
        column = []
        for entry in self._sort_list:
            row = self[entry._key]
            column.append(row.get(col, Table.NULL_VAL))
        return column

    def get_headers(self):
        """ Returns an OrderedSet """
        return self._headers

    def get_row_as_list(self, row_header, as_txt=False):
        if(row_header not in self):
            raise KeyError("Unable to find the row_header: {}\n"
                           "In Table object named: {}"
                           .format(row_header, self.name))

        if(as_txt):
            row_list = [str(row_header)]
        else:
            row_list = [row_header]

        for col_header in self._headers:
            if(as_txt):
                row_list.append(str(
                    self[row_header].get(col_header, Table.NULL_VAL)))
            else:
                row_list.append(
                    self[row_header].get(col_header, Table.NULL_VAL))

        return row_list

    def get_sort_list(self):
        """
        """
        if(self._sort_key is None):
            raise TypeError("The sort key is None. The sort key needs to be "
                            "set before calling the get_sort_list method.\n"
                            "Table: {}".format(self.name))
        return self._sort_list

    def rename_row(self, old_name, new_name):
        """
        """
        row_dict = self[old_name]
        del(self[old_name])
        self._sort_list.remove(old_name)
        self[new_name] = row_dict

    def set_sort_key(self, key, sort=True):
        """
        """
        if(key not in self._headers and key != "row_header"):
            raise KeyError("Attempt to set sort_key to a value that is not a "
                           "header.\n"
                           "Key: {}\n"
                           "Available headers: {}\n"
                           "Table name: {}"
                           .format(key, self._headers, self.name))

        self._sort_key = key

        for i, entry in enumerate(self._sort_list):
            if(self._sort_key == "row_header"):
                entry._val = entry._key
            else:
                row = self[entry._key]
                entry._val = row.get(self._sort_key, Table.NULL_VAL)
            self._sort_list[i] = entry

        if(sort is True):
            self._sort_list.sort()

    def set_sort_list(self, list):
        """
        """
        if(_sort_key is None):
            raise TypeError("The sort key is None. The sort key needs to be "
                            "set before calling the set_sort_list method.\n"
                            "Table: {}".format(self.name))
        self._sort_list = list

    def __iter__(self):
        """
        """
        self._iter_count = 0
        return self

    def __next__(self):
        """
        """
        if(self._iter_count >= len(self)):
            raise StopIteration
        entry = self._sort_list[self._iter_count]
        self._iter_count += 1
        return entry._key

    def __setitem__(self, row_header=None, col_val=None):
        """
            TODO
        """
        if(col_val is None):
            raise TypeError("Column key/value argument must be a dict, list or"
                            " tuple. Value was None.\n"
                            "Table: {}".format(self.name))

        if(row_header is None):
            row_header = self._row_header_count
            self._row_header_count += 1

        row_dict = self.get(row_header, Row(table=self))

        # If row is not new. Sort key has been added to the sort list
        # previously.
        sort_key_added = True
        if(not row_dict):
            # If row is new a sort key needs to be added to the sort list.
            sort_key_added = False
            if(self._sort_key == "row_header"):
                self._sort_list.append(row_header, row_header, ignore_len=True)
                sort_key_added = True

        # Several values must be added as a dict.
        if(isinstance(col_val, dict)):
            for col_name, val in col_val.items():
                if(col_name not in self._headers):
                    self.add_header(col_name)
                # Add sort key if needed
                if(self._sort_key == col_name and sort_key_added is False):
                    self._sort_list.append(row_header, val, ignore_len=True)
                    sort_key_added = True
                row_dict[col_name] = val
            # Add sort key if needed
            if(not sort_key_added):
                self._sort_list.append(row_header, Table.NULL_VAL,
                                       ignore_len=True)
            return super(Table, self).__setitem__(row_header, row_dict)

        # A single cell value in a table can be added using a list or tuple.
        if(isinstance(col_val, list) or isinstance(col_val, tuple)):
            if(len(col_val) == 2):
                if(col_val[0] not in self._headers):
                    self.add_header(col_val[0])
                row_dict[col_val[0]] = col_val[1]
                return super(Table, self).__setitem__(row_header, row_dict)
            else:
                raise ValueError("Unexpected number of values. Expected 2. "
                                 "Got: {}\n"
                                 "Input: {}\n"
                                 "Table: {}".format(len(col_val), col_val,
                                                    self.name))

        raise TypeError("Unexpected type. The value must be of the type Row, "
                        "list, or tuple.\n"
                        "Recieved type: {}\n"
                        "Table: {}\n".format(type(col_name), self.name))


class ColumnHeader():
    def __init__(self, table, header):
        """
        """
        table._col_header_count += 1
        self._header = header
        self._index = table._col_header_count

    def __cmp__(self, other):
        if(not isinstance(other, ColumnHeader)):
            raise TypeError("A ColumnHeader object cannot be compared to a "
                            "non-ColumnHeader object.\n"
                            "ColumnHeader _header: {} _index: {}\n"
                            "Other object was of type: {}"
                            .format(self._header, self._index, type(other)))
        return super(ColumnHeader, self._index).__cmp__(other._index)


class SortList(list):
    """
        Used together with the Table class, as a way to extract rows/data in a
        specific order defined by objects of the type SortList.

        Programmers are meant to sort this list in the desired order and store
        it in a Table class object which will then output rows/data in the
        order defined by the list.

        Implementations details:
        The SortList object contain SortListEntries containing key variables
        with unique values which define the order. And value variables which
        is the 'public' values used to sort on.
    """
    def __init__(self, container=None, unique_list=None, val_list=None):
        self.container = None
        if(container is not None):
            self.container = container

        if(unique_list is None and val_list is None):
            return
        if(len(unique_list) != len(val_list)):
            raise IndexError("A SortList object was initiated with two lists "
                             "of different sizes.\n"
                             "len(unique_list) {} != {} len(val_list)"
                             .format(len(unique_list), len(val_list)))

        for i, val in enumerate(val_list):
            self.append(unique_list[i], val)

    def append(self, key, val, ignore_len=False):
        entry = SortListEntry(key, val)
        if(self.container is not None and ignore_len is False):
            if(len(self.container) != (len(self) + 1)):
                raise IndexError("Attempt to add an entry to SortList which "
                                 "would cause the list to be longer than the "
                                 "affiliated SortList container.")
        return super(SortList, self).append(entry)

    def remove(self, id):
        """
        """
        for i, entry in enumerate(self):
            if(id == entry._key):
                del(self[i])

    def __repr__(self):
        outstr_list = []
        for entry in self:
            outstr_list.append("{}\t{}".format(entry._key, entry._val))
        return "\n".join(outstr_list)


class SortListEntry():
    """
    """
    def __init__(self, key, val):
        self._key = key
        self._val = val

    def __cmp__(self, other):
        if(not isinstance(other, SortListEntry)):
            raise TypeError("A SortListEntry object cannot be compared to a "
                            "non-SortListEntry object.\n"
                            "SortListEntry key: {} val: {}\n"
                            "Other object was of type: {}"
                            .format(self._key, self._val, type(other)))

        sort_val = str(self._val).lower()
        sort_val_other = str(other._val).lower()
        return sort_val.__cmp__(sort_val_other)

    def __repr__(self):
        return "{},{}".format(self._key, self._val)

    def __lt__(self, other):
        sort_val = str(self._val).lower()
        sort_val_other = str(other._val).lower()
        return sort_val.__lt__(sort_val_other)

    def __le__(self, other):
        sort_val = str(self._val).lower()
        sort_val_other = str(other._val).lower()
        return sort_val.__le__(sort_val_other)

    def __gt__(self, other):
        sort_val = str(self._val).lower()
        sort_val_other = str(other._val).lower()
        return sort_val.__gt__(sort_val_other)

    def __ge__(self, other):
        sort_val = str(self._val).lower()
        sort_val_other = str(other._val).lower()
        return sort_val.__ge__(sort_val_other)

    def __eq__(self, other):
        sort_val = str(self._val).lower()
        sort_val_other = str(other._val).lower()
        return sort_val.__eq__(sort_val_other)

    def __ne__(self, other):
        sort_val = str(self._val).lower()
        sort_val_other = str(other._val).lower()
        return sort_val.__ne__(sort_val_other)



#!/usr/bin/env python3
#
# Program: 	PointFinder-3.0
# Author: 	Camilla Hundahl Johnsen
#
# Dependencies: KMA or NCBI-blast together with BioPython.

import os
import re
import sys
import math
import argparse
import subprocess
import random
import copy

from cgecore.blaster import Blaster
from cgecore.cgefinder import CGEFinder
from .phenotype2genotype.feature import ResMutation
from .phenotype2genotype.res_profile import PhenoDB

# TODO: delete deprecated dependencies:
from .output.table import TableResults


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class GeneListError(Exception):
    """ Raise when a specified gene is not found within the gene list.
    """
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(PanelNameError, self).__init__(message, *args)


class PointFinder(CGEFinder):

    GENE_FILE = "genes.txt"
    RNA_GENE_FILE = "RNA_genes.txt"
    PHENOTYPE_FILE = "phenotypes.txt"
    # Deprecated variable, use phenotype (kept for legacy dbs)
    RES_OVERVIEW_FILE = "resistens-overview.txt"
    PROMOTER_REGEX = r"(.+promoter)-size-(\d+)(?:bp)"

    def __init__(self, db_path, species, gene_list=None, ignore_indels=False,
                 ignore_stop_codons=False):
        """
        """
        self.species = species
        self.specie_path = db_path
        self.RNA_gene_list = []
        self.ignore_indels = ignore_indels
        self.ignore_stop_codons = ignore_stop_codons

        self.gene_list = PointFinder.get_file_content(
            f"{self.specie_path}/{self.GENE_FILE}")

        self.promoters = PointFinder.load_promoters(self.gene_list)

        if os.path.isfile(f"{self.specie_path}/{self.RNA_GENE_FILE}"):
            self.RNA_gene_list = PointFinder.get_file_content(
                f"{self.specie_path}/{self.RNA_GENE_FILE}")

        # Creat user defined gene_list if given
        if(gene_list is not None):
            self.gene_list = self.get_user_defined_gene_list(gene_list)

        # Depends on database format, current or legacy
        if os.path.exists(f"{self.specie_path}/{self.PHENOTYPE_FILE}"):
            self.known_mutations, self.drug_genes, self.known_stop_codon = (
                self.get_db_mutations(
                    f"{self.specie_path}/{self.PHENOTYPE_FILE}",
                    self.gene_list))
        elif os.path.exists(f"{self.specie_path}/{self.RES_OVERVIEW_FILE}"):
            self.known_mutations, self.drug_genes, self.known_stop_codon = (
                self.get_db_old_mutations(
                    f"{self.specie_path}/{self.RES_OVERVIEW_FILE}",
                    self.gene_list))
        else:
            raise OSError("The pointfinder db does not have neither the"
                          f"'{self.PHENOTYPE_FILE}' file (current database) or"
                          "the '{self.RES_OVERVIEW_FILE}' file "
                          "(legacy database)")

    @staticmethod
    def load_promoters(gene_file_list):
        promoters = {}

        for gene_filename in gene_file_list:
            promoter_match = re.search(PointFinder.PROMOTER_REGEX, gene_filename)
            if promoter_match:
                promoter_key = promoter_match.group(1)
                promoter_length = promoter_match.group(2)
                promoters[promoter_key] = int(promoter_length)
        
        return promoters

    def get_user_defined_gene_list(self, gene_list):
        genes_specified = []
        for gene in gene_list:
            # Check that the genes are valid
            if gene not in self.gene_list:
                raise(GeneListError(
                    "Input Error: Specified gene not recognised "
                    "(%s)\nChoose one or more of the following genes:"
                    "\n%s" % (gene, "\n".join(self.gene_list))))
            genes_specified.append(gene)
        # Change the gene_list to the user defined gene_list
        return genes_specified

    #DELETE
    def old_results_to_standard_output(self, result, software, version,
                                       run_date, run_cmd, id,
                                       unknown_mut=False, tableresults=None):
        """
        """
        std_results = TableResults(software, version, run_date, run_cmd, id)
        headers = [
            "template_name",
            "template_length",
            "aln_length",
            "aln_identity",
            "aln_gaps",
            "aln_template_string",
            "aln_query_string",
            "aln_homology_string",
            "query_id",
            "query_start_pos",
            "query_end_pos",
            "template_variant",
            "query_depth",
            "blast_eval",
            "blast_bitscore",
            "pval",
            "software_score",
            "mutation",
            "ref_codon",
            "alt_codon",
            "ref_aa",
            "alt_aa",
            "insertion",
            "deletion"
        ]
        for db_name, db in result.items():
            if(db_name == "excluded"):
                continue

            if(db == "No hit found"):
                continue

            ####Added to solve PointFinder####
            if(isinstance(db, str)):
                if db.startswith("Gene found with coverage"):
                    continue
            ####

            # Start writing output string (to HTML tab file)
            gene_name = PhenoDB.if_promoter_rename(db_name)

            # Find and save mis_matches in gene
            sbjct_start = db["sbjct_start"]
            sbjct_seq = db["sbjct_string"]
            qry_seq = db["query_string"]
            db["mis_matches"] = self.find_mismatches(gene=db_name,
                                                     sbjct_start=sbjct_start,
                                                     sbjct_seq=sbjct_seq,
                                                     qry_seq=qry_seq)

            known_muts = []
            unknown_muts = []
            if(len(db["mis_matches"]) > 0):
                known_muts, unknown_muts = self.get_mutations(
                    db_name, gene_name, db['mis_matches'], True, db)

            # No mutations found
            if(not (unknown_muts or known_muts)):
                continue

            std_results.add_table(db_name)
            std_db = std_results.long[db_name]
            std_db.add_headers(headers)
            std_db.lock_headers = True

            for mut in known_muts:
                unique_id = mut.unique_id
                std_db[unique_id] = {
                    "template_name": db_name,
                    "template_length": db["sbjct_length"],
                    "aln_length": db["HSP_length"],
                    "aln_identity": db["perc_ident"],
                    "aln_gaps": db["gaps"],
                    "aln_template_string": db["sbjct_string"],
                    "aln_query_string": db["query_string"],
                    "aln_homology_string": db["homo_string"],
                    "query_id": db["contig_name"],
                    "query_start_pos": mut.pos,
                    "query_end_pos": mut.end,
                    "query_depth": db.get("depth", "NA"),
                    "pval": db.get("p_value", "NA"),
                    "software_score": db["cal_score"],
                    "mutation": mut.mut_string_short,
                    "ref_codon": mut.ref_codon,
                    "alt_codon": mut.mut_codon,
                    "ref_aa": mut.ref_aa,
                    "alt_aa": mut.mut_aa,
                    "insertion": mut.insertion,
                    "deletion": mut.deletion
                }

        return std_results

    @staticmethod
    def get_db_names(db_root_path):
        out_lst = []
        for entry in os.scandir(db_root_path):
            if not entry.name.startswith('.') and entry.is_dir():
                out_lst.append(entry.name)
        return tuple(out_lst)

    def results_to_str(self, res_type, results, unknown_flag, min_cov,
                       perc_iden):
        # Initiate output stings with headers
        gene_list = ', '.join(self.gene_list)
        output_strings = [
            "Mutation\tNucleotide change\tAmino acid change\tResistance\tPMID",
            "Chromosomal point mutations - Results\nSpecies: %s\nGenes: %s\n"
            "Mapping methode: %s\n\n\nKnown Mutations\n" % (self.species, gene_list, res_type), ""
        ]
        # Get all drug names and add header of all drugs to prediction file
        drug_lst = [drug for drug in self.drug_genes.keys()]
        output_strings[2] = "\t".join(drug_lst) + "\n"

        # Define variables to write temporary output into
        total_unknown_str = ""
        unique_drug_list = []
        excluded_hits = {}

        if res_type == PointFinder.TYPE_BLAST:
            # Patch together partial sequence hits (BLASTER does not do this)
            # and removes dummy_hit_id
            GENES = self.find_best_seqs(results, min_cov)

        else:
            # TODO: Some databases are either genes or species,
            #       depending on the method applied (BLAST/KMA).
            GENES = results[self.species]

            # If not hits found insert genes not found
            if GENES == 'No hit found':
                GENES = {}

            # KMA only gives the genes found, however all genes should be
            # included
            str_keys = str(GENES.keys())
            for gene in self.gene_list:
                if gene not in str_keys:
                    GENES[gene] = 'No hit found'

            # Included excluded
            GENES['excluded'] = results['excluded']
            for gene, entry in GENES.items():
                if entry != 'No hit found' and gene != 'excluded':
                    entry['mis_matches'] = self.find_mismatches(gene,
                                            entry['sbjct_start'],
                                            entry['sbjct_string'],
                                            entry['query_string'])

        for gene, hit in GENES.items():
            # Start writing output string (to HTML tab file)
            gene_name = gene.split("_")[0]  # Not perfeft can differ from specific mutation
            gene_name = PhenoDB.if_promoter_gene_rename(gene_name)
            output_strings[1] += "\n%s\n" % (gene_name)

            # Ignore excluded genes
            # TODO: Should all not found genes be moved to this dict?
            if gene == "excluded":
                continue

            # Write exclusion reason for genes not found
            if isinstance(GENES[gene], str):
                output_strings[1] += GENES[gene] + "\n"
                continue

            if hit['perc_ident'] < perc_iden and hit['perc_coverage'] < min_cov:
                GENES[gene] = ('Gene found with identity, %s, below minimum '
                               'identity threshold: %s. And with coverage, %s,'
                               ' below minimum coverage threshold: %s.'
                                % (hit['perc_ident'], perc_iden,
                                   hit['perc_coverage'], min_cov))
                output_strings[1] += GENES[gene] + "\n"
                continue

            if hit['perc_ident'] < perc_iden:
                GENES[gene] = ('Gene found with identity, %s, below minimum '
                               'identity threshold: %s' % (hit['perc_ident'],
                                                                   perc_iden))
                output_strings[1] += GENES[gene] + "\n"
                continue

            if hit['perc_coverage'] < min_cov:
                # Gene not found above given coverage
                GENES[gene] = ('Gene found with coverage, %s, below minimum '
                               'coverage threshold: %s' % (hit['perc_coverage'],
                                                                   min_cov))
                output_strings[1] += GENES[gene] + "\n"
                continue

            sbjct_start = hit['sbjct_start']
            sbjct_seq = hit['sbjct_string']
            qry_seq = hit['query_string']
            mis_matches = []
            if res_type == PointFinder.TYPE_BLAST:
                for subhit in hit['hits']:
                    if len(hit['hits'][subhit]['mis_matches']) > 0:
                        mis_matches.extend(hit['hits'][subhit]['mis_matches'])
            else:
                if "mis_matches" in hit:
                    mis_matches.extend(hit['mis_matches'])
            # Find and save mis_matches in gene
            hit['mis_matches'] = self.find_mismatches(gene, sbjct_start,
                                                      sbjct_seq, qry_seq)

            # Check if no mutations was found
            if len(mis_matches) < 1:
                output_strings[1] += ("No mutations found in {}\n"
                                      .format(gene_name))
            else:
                # Write mutations found to output file
                total_unknown_str += "\n%s\n" % (gene_name)

                str_tuple = self.mut2str(gene, gene_name, mis_matches,
                                         unknown_flag, hit)

                all_results = str_tuple[0]
                total_known = str_tuple[1]
                total_unknown = str_tuple[2]
                drug_list = str_tuple[3]

                # Add results to output strings
                if(all_results):
                    output_strings[0] += "\n" + all_results
                output_strings[1] += total_known + "\n"

                # Add unknown mutations the total results of
                # unknown mutations
                total_unknown_str += total_unknown + "\n"

                # Add drugs to druglist
                unique_drug_list += [drug.lower() for drug in drug_list]

        if unknown_flag is True:
            output_strings[1] += ("\n\nUnknown Mutations \n"
                                  + total_unknown_str)

        # Make Resistance Prediction output

        # Go throug all drugs in the database and see if prediction
        # can be called.
        pred_output = []
        for drug in drug_lst:
            drug = drug.lower()
            # Check if resistance to drug was found
            if drug in unique_drug_list:
                pred_output.append("1")
            else:
                # Check at all genes associated with the drug
                # resistance where found
                all_genes_found = True
                for gene in self.drug_genes[drug]:
                    if gene not in GENES:
                        all_genes_found = False

                if all_genes_found is False:
                    pred_output.append("?")
                else:
                    pred_output.append("0")

        output_strings[2] += "\t".join(pred_output) + "\n"

        return output_strings

    def write_results(self, out_path, result, res_type, unknown_flag, min_cov,
                      perc_iden):
        result_str = self.results_to_str(res_type=res_type,
                                         results=result,
                                         unknown_flag=unknown_flag,
                                         min_cov=min_cov,
                                         perc_iden=perc_iden)

        with open(out_path + "/PointFinder_results.txt", "w") as fh:
            fh.write(result_str[0])
        with open(out_path + "/PointFinder_table.txt", "w") as fh:
            fh.write(result_str[1])

    @staticmethod
    def discard_unknown_muts(results_pnt, phenodb, method):

        if method == PointFinder.TYPE_BLAST:
            for gene_hits in results_pnt.values():
                try:
                    hits = gene_hits.get("hits", {})

                    # Should only be one hit (the best)
                    for hit_key, entry in hits.items():
                        known_muts = PointFinder._get_known_mis_matches(
                            entry['sbjct_header'], entry["mis_matches"], phenodb)
                        entry["mis_matches"] = known_muts
                        gene_hits["mis_matches"] = known_muts
                except AttributeError:
                    pass
        else: #method == KMA
            for key, gene_hits in results_pnt.items():
                if key == 'excluded' or isinstance(gene_hits, str):
                    continue
                known_muts = PointFinder._get_known_mis_matches(
                    gene_hits['sbjct_header'], gene_hits["mis_matches"], phenodb)
                gene_hits["mis_matches"] = known_muts

        return results_pnt

    @staticmethod
    def _get_known_mis_matches(entry_key, mis_matches, phenodb):

        gene_db_id = entry_key.replace("_", ";;")
        known = []
        for mis_match in mis_matches:
            if mis_match[0] != 'sub':
                mis_match_key_nuc = (f"{gene_db_id}_{mis_match[2]}"
                                     f"_{mis_match[3].lower()}")
                mis_match_key_aa = (f"{gene_db_id}_{mis_match[1]}"
                                    f"_{mis_match[-1].lower()}")
            elif mis_match[4].startswith('r.'):  # RNA substitution
                mis_match_key_aa = ""
                mis_match_key_nuc = (f"{gene_db_id}_{mis_match[2]}"
                                     f"_{mis_match[3].lower()}")
            elif mis_match[4].startswith('n.'):  # non-coding DNA substitution
                mis_match_key_aa = ""
                mis_match_key_nuc = (f"{gene_db_id}_{mis_match[2]}"
                                     f"_{mis_match[3].lower()}")
            else:
                mis_match_key_aa = (f"{gene_db_id}_{mis_match[2]}"
                                    f"_{mis_match[-1].lower()}")
                mis_match_key_nuc = ""

            if phenodb.mut_type_is_defined:
                mis_match_key_nuc = mis_match_key_nuc + '_NUC'
                mis_match_key_aa = mis_match_key_aa + '_AA'

            if (mis_match_key_nuc in phenodb
                    or mis_match_key_aa in phenodb):
                known.append(mis_match)
        return known

    @staticmethod
    def discard_from_dict(in_dict, wanted_list):
        """
            Takes a dict and a list of keys.
            Returns a dict containing only the keys from the list.
        """
        out_dict = dict()
        for key, val in in_dict.items():
            if(key in wanted_list):
                out_dict[key] = val
        return out_dict

    def blast(self, inputfile, out_path, min_cov=0.6, threshold=0.9,
              blast="blastn", cut_off=True):
        """
        """
        blast_run = Blaster(inputfile=inputfile, databases=self.gene_list,
                            db_path=self.specie_path, out_path=out_path,
                            min_cov=min_cov, threshold=threshold, blast=blast,
                            cut_off=cut_off)

        self.blast_results = blast_run.results  # TODO Is this used?
        return blast_run

    @staticmethod
    def get_db_old_mutations(mut_db_path, gene_list):
        """
        This function opens the file resistens-overview.txt, and reads
        the content into a dict of dicts. The dict will contain
        information about all known mutations given in the database.
        This dict is returned.
        This function reads the old pointfinder database
        """

        # Initiate variables
        known_mutations = dict()
        known_stop_codon = dict()
        drug_genes = dict()
        indelflag = False
        stopcodonflag = False

        # Go throug mutation file line by line

        with open(mut_db_path, "r") as fh:
            mutdb_file = fh.readlines()
        mutdb_file = [line.strip() for line in mutdb_file]

        for line in mutdb_file:
            # Ignore headers and check where the indel section starts
            if line.startswith("#"):
                if "indel" in line.lower():
                    indelflag = True
                elif "stop codon" in line.lower():
                    stopcodonflag = True
                else:
                    stopcodonflag = False
                continue

            mutation = [data.strip() for data in line.split("\t")]

            gene_ID = mutation[0]

            # Only consider mutations in genes found in the gene list
            if gene_ID in gene_list:
                gene_name = mutation[1]
                mut_pos = int(mutation[2])
                ref_codon = mutation[3]                     # Ref_nuc (1 or 3?)
                ref_aa = mutation[4]                        # Ref_codon
                alt_aa = mutation[5].split(",")             # Res_codon
                res_drug = mutation[6].replace("\t", " ")
                pmid = mutation[7].split(",")

                # Check if stop codons are predictive for resistance
                if stopcodonflag is True:
                    if gene_ID not in known_stop_codon:
                        known_stop_codon[gene_ID] = {"pos": [],
                                                     "drug": res_drug}
                    known_stop_codon[gene_ID]["pos"].append(mut_pos)

                # Add genes associated with drug resistance to drug_genes dict
                drug_lst = res_drug.split(",")
                drug_lst = [d.strip().lower() for d in drug_lst]
                for drug in drug_lst:
                    if drug not in drug_genes:
                        drug_genes[drug] = []
                    if gene_ID not in drug_genes[drug]:
                        drug_genes[drug].append(gene_ID)

                # Initiate empty dict to store relevant mutation information
                mut_info = dict()

                # Save need mutation info with pmid cooresponding to the amino
                # acid change
                for i in range(len(alt_aa)):
                    try:
                        mut_info[alt_aa[i]] = {"gene_name": gene_name,
                                               "drug": res_drug,
                                               "pmid": pmid[i]}
                    except IndexError:
                        mut_info[alt_aa[i]] = {"gene_name": gene_name,
                                               "drug": res_drug,
                                               "pmid": "-"}

                # Add all possible types of mutations to the dict
                if gene_ID not in known_mutations:
                    known_mutations[gene_ID] = {"sub": dict(), "ins": dict(),
                                                "del": dict()}
                # Check for the type of mutation
                if indelflag is False:
                    mutation_type = "sub"
                else:
                    mutation_type = ref_aa

                # Save mutations positions with required information given in
                # mut_info
                if mut_pos not in known_mutations[gene_ID][mutation_type]:
                    known_mutations[gene_ID][mutation_type][mut_pos] = dict()
                for amino in alt_aa:
                    if (amino in known_mutations[gene_ID][mutation_type]
                                                [mut_pos]):
                        stored_mut_info = (known_mutations[gene_ID]
                                                          [mutation_type]
                                                          [mut_pos][amino])
                        if stored_mut_info["drug"] != mut_info[amino]["drug"]:
                            stored_mut_info["drug"] += "," + (mut_info[amino]
                                                                      ["drug"])
                        if stored_mut_info["pmid"] != mut_info[amino]["pmid"]:
                            stored_mut_info["pmid"] += ";" + (mut_info[amino]
                                                                      ["pmid"])

                        (known_mutations[gene_ID][mutation_type]
                                        [mut_pos][amino]) = stored_mut_info
                    else:
                        (known_mutations[gene_ID][mutation_type]
                                        [mut_pos][amino]) = mut_info[amino]

        # Check that all genes in the gene list has known mutations
        for gene in gene_list:
            if gene not in known_mutations:
                known_mutations[gene] = {"sub": dict(), "ins": dict(),
                                         "del": dict()}

        return known_mutations, drug_genes, known_stop_codon

    @staticmethod
    def get_db_mutations(mut_db_path, gene_list):
        """
        This function opens the file self.PHENOTYPE_FILE, and reads
        the content into a dict of dicts. The dict will contain
        information about all known mutations given in the database.
        This dict is returned.
        """

        # Initiate variables
        known_mutations = dict()
        known_stop_codon = dict()
        drug_genes = dict()
        indelflag = False
        stopcodonflag = False

        # Go throug mutation file line by line

        with open(mut_db_path, "r") as fh:
            mutdb_file = fh.readlines()
        mutdb_file = [line.strip() for line in mutdb_file]

        for line in mutdb_file:
            # Ignore headers and check where the indel section starts
            if line.startswith("#"):
                if "indel" in line.lower():
                    indelflag = True
                elif "stop codon" in line.lower():
                    stopcodonflag = True
                else:
                    stopcodonflag = False
                continue

            mutation = [data.strip() for data in line.split("\t")]
            gene_ID = mutation[0]
            gene_name = str(mutation[0].split("_")[0])
            # Only consider mutations in genes found in the gene list
            if gene_name in gene_list:
                gene_name = PhenoDB.if_promoter_gene_rename(gene_name)
                mut_pos = int(mutation[3])
                ref_codon = mutation[4]                     # Ref_nuc (1 or 3?)
                ref_aa = mutation[5]                        # Ref_codon
                alt_aa = mutation[6].split(",")             # Res_codon
                res_drug = mutation[8].replace("\t", " ")
                pmid = mutation[9].split(",")

                # Add genes associated with drug resistance to drug_genes dict
                drug_lst = res_drug.split(",")
                drug_lst = [d.strip().lower() for d in drug_lst]
                for drug in drug_lst:
                    if drug not in drug_genes:
                        drug_genes[drug] = []
                    if gene_name not in drug_genes[drug]:
                        drug_genes[drug].append(gene_name)

                # Check if stop codons are predictive for resistance
                if stopcodonflag is True:
                    if gene_name not in known_stop_codon:
                        known_stop_codon[gene_name] = {"pos": [],
                                                       "drug": res_drug}
                    known_stop_codon[gene_name]["pos"].append(mut_pos)

                # Initiate empty dict to store relevant mutation information
                mut_info = dict()

                # Save needed mutation info with pmid cooresponding to the
                # amino acid change
                for i in range(len(alt_aa)):
                    try:
                        mut_info[alt_aa[i]] = {"gene_name": gene_name,
                                               "drug": drug_lst,
                                               "pmid": [pmid[i]]}
                    except IndexError:
                        mut_info[alt_aa[i]] = {"gene_name": gene_name,
                                               "drug": drug_lst,
                                               "pmid": "-"}

                # Add all possible types of mutations to the dict
                if gene_name not in known_mutations:
                    known_mutations[gene_name] = {"sub": dict(), "ins": dict(),
                                                  "del": dict()}
                # Check for the type of mutation
                if indelflag is False:
                    mutation_type = "sub"
                else:
                    mutation_type = ref_aa

                # Save mutations positions with required information given in
                # mut_info
                if mut_pos not in known_mutations[gene_name][mutation_type]:
                    known_mutations[gene_name][mutation_type][mut_pos] = dict()

                for amino in alt_aa:
                    stored_info = (known_mutations[gene_name][mutation_type]
                                   [mut_pos].get(amino, False))
                    if stored_info:
                        merged_info = PointFinder.merge_mut_info(stored_info,
                                                                 amino,
                                                                 mut_info)
                        (known_mutations[gene_name][mutation_type]
                                        [mut_pos][amino]) = merged_info
                    else:
                        (known_mutations[gene_name][mutation_type]
                                        [mut_pos][amino]) = mut_info[amino]

        # Check that all genes in the gene list has known mutations
        for gene in gene_list:
            if gene not in known_mutations:
                known_mutations[gene] = {"sub": dict(), "ins": dict(),
                                         "del": dict()}
        return known_mutations, drug_genes, known_stop_codon

    @staticmethod
    def merge_mut_info(stored_info, amino, mut_info):
        for drug in mut_info[amino]["drug"]:
            if drug not in stored_info["drug"]:
                stored_info["drug"].append(drug)

        for pmid in mut_info[amino]["pmid"]:
            if pmid not in stored_info["pmid"]:
                stored_info["pmid"].append(pmid)

        return stored_info

    def find_best_seqs(self, org_blast_results, min_cov):
        """
        This function finds gene sequences with the largest coverage based on
        the blast results. If more hits covering different sequences parts
        exists it concatinates partial gene hits into one hit.
        If different overlap sequences occurs they are saved in the list
        alternative_overlaps. The function returns a new results dict where
        each gene has an inner dict with hit information corresponding to
        the new concatinated hit. If no gene is found the value is a string
        with info of why the gene wasn't found.
        """

        # Very inefficient to create a deep copy but safest option until
        # pointfinder code gets rewritten.
        blast_results = copy.deepcopy(org_blast_results)

        GENES = dict()
        for gene, hits in blast_results.items():

            # Save all hits in the list 'hits_found'
            hits_found = []
            GENES[gene] = {}

            # Check for gene has any hits in blast results
            if gene == 'excluded':
                GENES[gene] = hits
                continue
            elif isinstance(hits, dict) and len(hits) > 0:
                GENES[gene]['found'] = 'partially'
                # IT is possible only this is the used part?
                GENES[gene]['hits'] = hits
                for hit_ in hits:
                    GENES[gene]['hits'][hit_]['mis_matches'] = self.find_mismatches(
                        gene=gene,
                        sbjct_start=GENES[gene]['hits'][hit_]["sbjct_start"],
                        sbjct_seq=GENES[gene]['hits'][hit_]["sbjct_string"],
                        qry_seq=GENES[gene]['hits'][hit_]["query_string"])
            else:
                # Gene not found! go to next gene
                GENES[gene] = 'No hit found'
                continue

            # Check coverage for each hit, patch together partial genes hits
            for hit_id, hit in hits.items():

                # Save hits start and end positions in subject, total subject
                # len, and subject and query sequences of each hit
                hits_found += [(hit['sbjct_start'], hit['sbjct_end'],
                                hit['sbjct_string'], hit['query_string'],
                                hit['sbjct_length'], hit['homo_string'],
                                hit_id)]

                # If coverage is 100% change found to yes
                hit_coverage = hit['perc_coverage']
                if hit_coverage == 1.0:
                    GENES[gene]['found'] = 'yes'

            # Sort positions found
            hits_found = sorted(hits_found, key=lambda x: x[0])

            # Find best hit by concatenating sequences if more hits exist

            # Get information from the first hit found
            all_start = hits_found[0][0]
            current_end = hits_found[0][1]
            final_sbjct = hits_found[0][2]
            final_qry = hits_found[0][3]
            sbjct_len = hits_found[0][4]
            final_homol = hits_found[0][5]
            first_hit_id = hits_found[0][6]

            alternative_overlaps = []
            contigs = [hit['contig_name']]
            scores = [str(hit['cal_score'])]

            # Check if more than one hit was found within the same gene
            for i in range(len(hits_found) - 1):

                # Save information from previous hit
                pre_block_start = hits_found[i][0]
                pre_block_end = hits_found[i][1]
                pre_sbjct = hits_found[i][2]
                pre_qry = hits_found[i][3]
                pre_homol = hits_found[i][5]
                pre_id = hits_found[i][6]

                # Save information from next hit
                next_block_start = hits_found[i + 1][0]
                next_block_end = hits_found[i + 1][1]
                next_sbjct = hits_found[i + 1][2]
                next_qry = hits_found[i + 1][3]
                next_homol = hits_found[i + 1][5]
                next_id = hits_found[i + 1][6]

                contigs.append(hits[next_id]['contig_name'])
                scores.append(str(hits[next_id]['cal_score']))

                # Check for overlapping sequences, collaps them and save
                # alternative overlaps if any
                if next_block_start <= current_end:

                    # Find overlap start and take gaps into account
                    pos_count = 0
                    overlap_pos = pre_block_start
                    for i in range(len(pre_sbjct)):

                        # Stop loop if overlap_start position is reached
                        if overlap_pos == next_block_start:
                            overlap_start = pos_count
                            break
                        if pre_sbjct[i] != "-":
                            overlap_pos += 1
                        pos_count += 1

                    # Find overlap len and add next sequence to final sequence
                    if len(pre_sbjct[overlap_start:]) > len(next_sbjct):
                        #  <--------->
                        #     <--->
                        overlap_len = len(next_sbjct)
                        overlap_end_pos = next_block_end
                    else:
                        #  <--------->
                        #        <--------->
                        overlap_len = len(pre_sbjct[overlap_start:])
                        overlap_end_pos = pre_block_end

                        # Update current end
                        current_end = next_block_end

                        # Use the entire previous sequence and add the last
                        # part of the next sequence
                        final_sbjct += next_sbjct[overlap_len:]
                        final_qry += next_qry[overlap_len:]
                        final_homol += next_homol[overlap_len:]

                    # Find query overlap sequences
                    pre_qry_overlap = pre_qry[overlap_start: (overlap_start
                                                              + overlap_len)]
                    next_qry_overlap = next_qry[:overlap_len]
                    sbjct_overlap = next_sbjct[:overlap_len]

                    # If alternative query overlap excist save it
                    if pre_qry_overlap != next_qry_overlap:
                        # Save alternative overlaps
                        alternative_overlaps += [(next_block_start,
                                                  overlap_end_pos,
                                                  sbjct_overlap,
                                                  next_qry_overlap)]

                elif next_block_start > current_end:
                    #  <------->
                    #              <------->
                    gap_size = next_block_start - current_end - 1
                    final_qry += "N" * gap_size
                    final_sbjct += "N" * gap_size
                    final_homol += "-" * gap_size
                    current_end = next_block_end
                    final_sbjct += next_sbjct
                    final_qry += next_qry
                    final_homol += next_homol

            # Calculate new coverage
            no_call = final_qry.upper().count("N")
            coverage = ((current_end - all_start + 1 - no_call)
                        / float(sbjct_len))

            # Calculate identity
            equal = 0
            not_equal = 0
            for i in range(len(final_qry)):
                if final_qry[i].upper() != "N":
                    if final_qry[i].upper() == final_sbjct[i].upper():
                        equal += 1
                    else:
                        not_equal += 1
            identity = equal / float(equal + not_equal)
            if coverage >= min_cov:
                GENES[gene]['perc_coverage'] = coverage
                GENES[gene]['perc_ident'] = identity
                GENES[gene]['sbjct_string'] = final_sbjct
                GENES[gene]['query_string'] = final_qry
                GENES[gene]['homo_string'] = final_homol
                GENES[gene]['contig_name'] = ", ".join(contigs)
                GENES[gene]['HSP_length'] = len(final_qry)
                GENES[gene]['sbjct_start'] = all_start
                GENES[gene]['sbjct_end'] = current_end
                GENES[gene]['sbjct_length'] = sbjct_len
                GENES[gene]['cal_score'] = ", ".join(scores)
                GENES[gene]['gaps'] = no_call
                GENES[gene]['alternative_overlaps'] = alternative_overlaps
                mismatches = self.find_mismatches(gene=gene,
                                                  sbjct_start=all_start,
                                                  sbjct_seq=final_sbjct,
                                                  qry_seq=final_qry)
                GENES[gene]['mis_matches'] = mismatches

            else:
                # Gene not found above given coverage
                GENES[gene] = ('Gene found with coverage, %f, below '
                               'minimum coverage threshold: %s' % (coverage,
                                                                   min_cov))
        return GENES

    def find_mismatches(self, gene, sbjct_start, sbjct_seq, qry_seq):
        """
        This function finds mis matches between two sequeces. Depending
        on the the sequence type either the function
        find_codon_mismatches or find_nucleotid_mismatches are called,
        if the sequences contains both a promoter and a coding region
        both functions are called. The function can also call itself if
        alternative overlaps is give. All found mismatches are returned
        """

        # Initiate the mis_matches list that will store all found mis matcehs
        mis_matches = []
        gene_name = gene.split('_')[0]

        # Find mis matches in RNA genes
        if gene_name in self.RNA_gene_list:
            mis_matches += PointFinder.find_nucleotid_mismatches(sbjct_start,
                                                                 sbjct_seq,
                                                                 qry_seq)
        else:
            # Check if the gene sequence is with a promoter
            promtr_len = self.promoters.get(gene_name, None)
            if promtr_len is not None:

                # Extract promoter sequence, while considering gaps
                # --------agt-->----
                #    ---->?
                if sbjct_start <= promtr_len:
                    # Find position in sbjct sequence where promoter ends
                    promtr_end = 0
                    nuc_count = sbjct_start - 1

                    for i in range(len(sbjct_seq)):
                        promtr_end += 1

                        if sbjct_seq[i] != "-":
                            nuc_count += 1

                        if nuc_count == promtr_len:
                            break

                    # Check if only a part of the promoter is found
                    # --------agt-->----
                    # ----
                    promtr_sbjct_start = -1
                    if nuc_count < promtr_len:
                        promtr_sbjct_start = nuc_count - promtr_len

                    # Get promoter part of subject and query
                    sbjct_promtr_seq = sbjct_seq[:promtr_end]
                    qry_promtr_seq = qry_seq[:promtr_end]

                    # For promoter part find nucleotide mis matches
                    mis_matches += PointFinder.find_nucleotid_mismatches(
                        promtr_sbjct_start, sbjct_promtr_seq, qry_promtr_seq,
                        promoter=True)

                    # Check if gene is also found
                    # --------agt-->----
                    #     -----------
                    if((sbjct_start + len(sbjct_seq.replace("-", "")))
                       > promtr_len):
                        sbjct_gene_seq = sbjct_seq[promtr_end:]
                        qry_gene_seq = qry_seq[promtr_end:]
                        sbjct_gene_start = 1

                        # Find mismatches in gene part
                        mis_matches += self.find_codon_mismatches(
                            sbjct_gene_start, sbjct_gene_seq, qry_gene_seq)

                # No promoter, only gene is found
                # --------agt-->----
                #            -----
                else:
                    sbjct_gene_start = sbjct_start - promtr_len

                    # Find mismatches in gene part
                    mis_matches += self.find_codon_mismatches(
                        sbjct_gene_start, sbjct_seq, qry_seq)

            else:
                # Find mismatches in gene
                mis_matches += self.find_codon_mismatches(
                    sbjct_start, sbjct_seq, qry_seq)

        return mis_matches

    @staticmethod
    def find_nucleotid_mismatches(sbjct_start, sbjct_seq, qry_seq,
                                  promoter=False):
        """
        This function takes two alligned sequence (subject and query),
        and the position on the subject where the alignment starts. The
        sequences are compared one nucleotide at a time. If mis matches
        are found they are saved. If a gap is found the function
        find_nuc_indel is called to find the entire indel and it is
        also saved into the list mis_matches. If promoter sequences are
        given as arguments, these are reversed the and the absolut
        value of the sequence position  used, but when mutations are
        saved the negative value and det reverse sequences are saved in
        mis_mathces.
        """

        # Initiate the mis_matches list that will store all found
        # mismatcehs
        mis_matches = []

        sbjct_start = abs(sbjct_start)

        # Set variables depending on promoter status
        seq_pos = 1
        factor = 1
        mut_prefix = "r."

        if promoter is True:
            seq_pos = sbjct_start
            factor = (-1)
            mut_prefix = "n."
            # Reverse promoter sequences
            sbjct_seq = sbjct_seq[::-1]
            qry_seq = qry_seq[::-1]

        # Go through sequences one nucleotide at a time
        shift = 0
        for index in range(seq_pos - 1, len(sbjct_seq)):
            mut_name = mut_prefix
            mut = ""
            # Shift index according to gaps
            i = index + shift

            # If the end of the sequence is reached, stop
            if i == len(sbjct_seq):
                break

            sbjct_nuc = sbjct_seq[i]
            qry_nuc = qry_seq[i]

            # Check for mis matches
            if sbjct_nuc.upper() != qry_nuc.upper():

                # check for insertions and deletions
                if sbjct_nuc == "-" or qry_nuc == "-":
                    if sbjct_nuc == "-":
                        mut = "ins"
                        indel_start_pos = (seq_pos - 1) * factor + sbjct_start
                        indel_end_pos = seq_pos * factor + sbjct_start
                        indel = PointFinder.find_nuc_indel(sbjct_seq[i:],
                                                           qry_seq[i:])
                    else:
                        mut = "del"
                        indel_start_pos = seq_pos * factor + (sbjct_start - 1)
                        indel = PointFinder.find_nuc_indel(qry_seq[i:],
                                                           sbjct_seq[i:])
                        indel_end_pos = (seq_pos + len(indel) - 1) * factor \
                                        + (sbjct_start - 1)
                        seq_pos += len(indel) - 1

                    # Shift the index to the end of the indel
                    shift += len(indel) - 1

                    # Write mutation name, depending on sequnce
                    if len(indel) == 1 and mut == "del":
                        mut_name += str(indel_start_pos) + mut + indel
                    else:
                        if promoter is True:
                            # Reverse the sequence and the start and
                            # end positions
                            indel = indel[::-1]
                            temp = indel_start_pos
                            indel_start_pos = indel_end_pos
                            indel_end_pos = temp

                        mut_name += (str(indel_start_pos) + "_"
                                     + str(indel_end_pos) + mut + indel)

                    mis_matches += [[mut, indel_start_pos, indel_end_pos,
                                    indel, mut_name, mut, indel]]

                # Check for substitutions mutations
                else:
                    mut = "sub"
                    pos = seq_pos * factor + (sbjct_start - 1)
                    mut_name += (str(pos)
                                 + sbjct_nuc + ">" + qry_nuc)

                    mis_matches += [[mut, pos, pos,
                                    qry_nuc, mut_name, sbjct_nuc, qry_nuc]]

            # Increment sequence position
            if mut != "ins":
                seq_pos += 1

        return mis_matches

    @staticmethod
    def find_nuc_indel(gapped_seq, indel_seq):
        """
        This function finds the entire indel missing in from a gapped
        sequence compared to the indel_seqeunce. It is assumes that the
        sequences start with the first position of the gap.
        """
        ref_indel = indel_seq[0]
        for j in range(1, len(gapped_seq)):
            if gapped_seq[j] == "-":
                ref_indel += indel_seq[j]
            else:
                break
        return ref_indel

    @staticmethod
    def aa(codon):
        """
        This function converts a codon to an amino acid. If the codon
        is not valid an error message is given, or else, the amino acid
        is returned.

        Potential future issue: If species are added that utilizes
                                alternative translation tables.
        """
        codon = codon.upper()
        aa = {"ATT": "I", "ATC": "I", "ATA": "I",
              "CTT": "L", "CTC": "L", "CTA": "L", "CTG": "L", "TTA": "L",
              "TTG": "L",
              "GTT": "V", "GTC": "V", "GTA": "V", "GTG": "V",
              "TTT": "F", "TTC": "F",
              "ATG": "M",
              "TGT": "C", "TGC": "C",
              "GCT": "A", "GCC": "A", "GCA": "A", "GCG": "A",
              "GGT": "G", "GGC": "G", "GGA": "G", "GGG": "G",
              "CCT": "P", "CCC": "P", "CCA": "P", "CCG": "P",
              "ACT": "T", "ACC": "T", "ACA": "T", "ACG": "T",
              "TCT": "S", "TCC": "S", "TCA": "S", "TCG": "S", "AGT": "S",
              "AGC": "S",
              "TAT": "Y", "TAC": "Y",
              "TGG": "W",
              "CAA": "Q", "CAG": "Q",
              "AAT": "N", "AAC": "N",
              "CAT": "H", "CAC": "H",
              "GAA": "E", "GAG": "E",
              "GAT": "D", "GAC": "D",
              "AAA": "K", "AAG": "K",
              "CGT": "R", "CGC": "R", "CGA": "R", "CGG": "R", "AGA": "R",
              "AGG": "R",
              "TAA": "*", "TAG": "*", "TGA": "*"}
        # Translate valid codon
        try:
            amino_a = aa[codon]
        except KeyError:
            amino_a = "?"
        return amino_a

    @staticmethod
    def get_codon(seq, codon_no, start_offset):
        """
        This function takes a sequece and a codon number and returns
        the codon found in the sequence at that position
        """
        seq = seq.replace("-", "")
        codon_start_pos = int(codon_no - 1) * 3 - start_offset
        codon = seq[codon_start_pos:codon_start_pos + 3]
        return codon

    @staticmethod
    def name_insertion(sbjct_seq, codon_no, sbjct_nucs, aa_alt, start_offset):
        """
        This function is used to name a insertion mutation based on the
        HGVS recommendation.
        """
        start_codon_no = codon_no - 1

        if len(sbjct_nucs) == 3:
            start_codon_no = codon_no

        start_codon = PointFinder.get_codon(sbjct_seq, start_codon_no,
                                            start_offset)
        end_codon = PointFinder.get_codon(sbjct_seq, codon_no, start_offset)
        pos_name = "p.%s%d_%s%dins%s" % (PointFinder.aa(start_codon),
                                         start_codon_no,
                                         PointFinder.aa(end_codon),
                                         codon_no, aa_alt)
        return pos_name

    @staticmethod
    def name_deletion(sbjct_seq, sbjct_rf_indel, sbjct_nucs, codon_no, aa_alt,
                      start_offset, mutation="del"):
        """
        This function is used to name a deletion mutation based on the
        HGVS recommendation. If combination of a deletion and an
        insertion is identified the argument 'mutation' is set to
        'delins' and the mutation name will indicate that the mutation
        is a delins mutation.
        """
        del_codon = PointFinder.get_codon(sbjct_seq, codon_no, start_offset)
        pos_name = "p.%s%d" % (PointFinder.aa(del_codon), codon_no)

        # This has been changed
        if len(sbjct_rf_indel) == 3 and mutation == "del":
            return pos_name + mutation

        end_codon_no = codon_no + math.ceil(len(sbjct_nucs) / 3) - 1
        end_codon = PointFinder.get_codon(sbjct_seq, end_codon_no,
                                          start_offset)
        pos_name += "_%s%d%s" % (PointFinder.aa(end_codon), end_codon_no,
                                 mutation)
        if mutation == "delins":
            pos_name += aa_alt
        return pos_name

    @staticmethod
    def name_indel_mutation(sbjct_seq, indel, sbjct_rf_indel, qry_rf_indel,
                            codon_no, mut, start_offset):
        """
        This function serves to name the individual mutations
        dependently on the type of the mutation.
        """
        # Get the subject and query sequences without gaps
        sbjct_nucs = sbjct_rf_indel.replace("-", "")
        qry_nucs = qry_rf_indel.replace("-", "")

        # Translate nucleotides to amino acids
        aa_ref = ""
        aa_alt = ""

        for i in range(0, len(sbjct_nucs), 3):
            aa_ref += PointFinder.aa(sbjct_nucs[i:i + 3])

        for i in range(0, len(qry_nucs), 3):
            aa_alt += PointFinder.aa(qry_nucs[i:i + 3])

        # Identify the gapped sequence
        if mut == "ins":
            gapped_seq = sbjct_rf_indel
        else:
            gapped_seq = qry_rf_indel

        gap_size = gapped_seq.count("-")

        # Write mutation names
        if gap_size < 3 and len(sbjct_nucs) == 3 and len(qry_nucs) == 3:
            # Write mutation name for substitution mutation
            mut_name = "p.%s%d%s" % (PointFinder.aa(sbjct_nucs), codon_no,
                                     PointFinder.aa(qry_nucs))
        elif len(gapped_seq) == gap_size:

            if mut == "ins":
                # Write mutation name for insertion mutation
                mut_name = PointFinder.name_insertion(sbjct_seq, codon_no,
                                                      sbjct_nucs, aa_alt,
                                                      start_offset)
                aa_ref = mut
            else:
                # Write mutation name for deletion mutation
                mut_name = PointFinder.name_deletion(sbjct_seq, sbjct_rf_indel,
                                                     sbjct_nucs, codon_no,
                                                     aa_alt, start_offset,
                                                     mutation="del")
                aa_alt = mut
        # Check for delins - mix of insertion and deletion
        else:
            # Write mutation name for a mixed insertion and deletion
            # mutation
            mut_name = PointFinder.name_deletion(sbjct_seq,
                                                 sbjct_rf_indel, sbjct_nucs,
                                                 codon_no, aa_alt,
                                                 start_offset,
                                                 mutation="delins")

        # Check for frameshift
        if gapped_seq.count("-") % 3 != 0:
            # Add the frameshift tag to mutation name
            mut_name += " - Frameshift"

        return mut_name, aa_ref, aa_alt

    @staticmethod
    def get_inframe_gap(seq, nucs_needed=3):
        """
        This funtion takes a sequnece starting with a gap or the
        complementary seqeuence to the gap, and the number of
        nucleotides that the seqeunce should contain in order to
        maintain the correct reading frame. The sequence is gone
        through and the number of non-gap characters are counted. When
        the number has reach the number of needed nucleotides the indel
        is returned. If the indel is a 'clean' insert or deletion that
        starts in the start of a codon and can be divided by 3, then
        only the gap is returned.
        """
        nuc_count = 0
        gap_indel = ""
        nucs = ""

        for i in range(len(seq)):
            # Check if the character is not a gap
            if seq[i] != "-":
                # Check if the indel is a 'clean'
                # i.e. if the insert or deletion starts at the first
                # nucleotide in the codon and can be divided by 3

                if(gap_indel.count("-") == len(gap_indel)
                   and gap_indel.count("-") >= 3 and len(gap_indel) != 0):
                    return gap_indel

                nuc_count += 1
            gap_indel += seq[i]

            # if the number of nucleotides in the indel equals the amount
            # needed for the indel, the indel is returned.
            if nuc_count == nucs_needed:
                return gap_indel

        # This will only happen if the gap is in the very end of a sequence
        return gap_indel

    @staticmethod
    def get_aa_indels(sbjct_seq, qry_seq, start_pos):
        """
        This function uses regex to find inserts and deletions in
        sequences given as arguments. A list of these indels are
        returned. The list includes, type of mutations(ins/del),
        subject codon no of found mutation, subject sequence position,
        insert/deletions nucleotide sequence, and the affected qry
        codon no.
        """

        seqs = [sbjct_seq, qry_seq]
        indels = []
        gap_obj = re.compile(r"-+")
        for i in range(len(seqs)):
            for match in gap_obj.finditer(seqs[i]):
                pos = int(match.start())
                gap = match.group()

                # Find position of the mutation corresponding to the
                # subject sequence
                sbj_pos = len(sbjct_seq[:pos].replace("-", "")) + start_pos

                # Get indel sequence and the affected sequences in
                # sbjct and qry in the reading frame
                indel = seqs[abs(i - 1)][pos:pos + len(gap)]

                # Find codon number for mutation
                codon_no = int(math.ceil((sbj_pos) / 3))
                ref_codon_end = int(math.ceil((sbj_pos + len(gap)) / 3))
                qry_pos = len(qry_seq[:pos].replace("-", "")) + start_pos
                qry_codon = int(math.ceil((qry_pos) / 3))

                if i == 0:
                    mut = "ins"
                else:
                    mut = "del"

                # indels.append([mut, codon_no, sbj_pos, indel, qry_codon])
                indels.append([mut, codon_no, ref_codon_end, indel, qry_codon])

        # Sort indels based on codon position and sequence position
        indels = sorted(indels, key=lambda x: (x[1], x[2]))

        return indels

    @staticmethod
    def _remove_insertions_in_seq(sbjct_seq, qry_seq):
        """
        Remove insertions in query sequence (ex.: assumed to be sequencing
        errors).
        """
        sbjct_seq_list = []
        qry_seq_list = []

        for i, char in enumerate(sbjct_seq):
            if char != '-':
                sbjct_seq_list.append(char)
                qry_seq_list.append(qry_seq[i])

        return ("".join(sbjct_seq_list), "".join(qry_seq_list))

    def find_codon_mismatches(self, sbjct_start, sbjct_seq, qry_seq):
        """
        This function takes two alligned sequence (subject and query),
        and the position on the subject where the alignment starts. The
        sequences are compared codon by codon. If a mismatch is
        found it is saved in 'mis_matches'. If a gap is found the
        function get_inframe_gap is used to find the indel sequence and
        keep the sequence in the correct reading frame. The function
        translate_indel is used to name indel mutations and translate
        the indels to amino acids. The function returns a list of tuples
        containing all needed information about the mutation in order
        to look it up in the database dict known mutation and the with
        the output files the the user.
        """
        mis_matches = []

        # Find start pos of first codon in frame, i_start
        codon_offset = (sbjct_start - 1) % 3
        i_start = 0

        if codon_offset != 0:
            i_start = 3 - codon_offset

        sbjct_start = sbjct_start + i_start

        # Set sequences in frame
        sbjct_seq = sbjct_seq[i_start:]
        qry_seq = qry_seq[i_start:]

        if self.ignore_indels:
            sbjct_seq, qry_seq = PointFinder._remove_insertions_in_seq(
                sbjct_seq, qry_seq)

        # Find codon number of the first codon in the sequence, start
        # at 0
        codon_no = int((sbjct_start - 1) / 3)  # 1,2,3 start on 0

        # s_shift and q_shift are used when gaps appears
        q_shift = 0
        s_shift = 0
        mut_no = 0

        if not self.ignore_indels:
            # Find inserts and deletions in sequence
            indel_no = 0
            indels = PointFinder.get_aa_indels(sbjct_seq, qry_seq, sbjct_start)

        # Go through sequence and save mutations when found
        for index in range(0, len(sbjct_seq), 3):
            # Count codon number
            codon_no += 1

            # Shift index according to gaps
            s_i = index + s_shift
            q_i = index + q_shift

            # Get codons
            sbjct_codon = sbjct_seq[s_i:s_i + 3]
            qry_codon = qry_seq[q_i:q_i + 3]

            if(len(sbjct_seq[s_i:].replace("-", ""))
               + len(qry_codon[q_i:].replace("-", "")) < 6):
                break

            # Check for mutations
            if sbjct_codon.upper() != qry_codon.upper():

                if(("-" in sbjct_codon or "-" in qry_codon)
                   and not self.ignore_indels):
                    # Check for codon insertions and deletions and
                    # frameshift mutations

                    # Get indel info
                    try:
                        indel_data = indels[indel_no]
                    except IndexError:
                        sys.exit("indel_data list is out of range, bug!")

                    mut = indel_data[0]
                    codon_no_indel = indel_data[1]
                    seq_pos = indel_data[2]
                    indel = indel_data[3]
                    indel_no += 1

                    # Get the affected sequence in frame for both for
                    # sbjct and qry
                    if mut == "ins":
                        sbjct_rf_indel = PointFinder.get_inframe_gap(
                            sbjct_seq[s_i:], 3)
                        qry_rf_indel = PointFinder.get_inframe_gap(
                            qry_seq[q_i:],
                            int(math.floor(len(sbjct_rf_indel) / 3) * 3))
                    else:
                        qry_rf_indel = PointFinder.get_inframe_gap(
                            qry_seq[q_i:], 3)
                        sbjct_rf_indel = PointFinder.get_inframe_gap(
                            sbjct_seq[s_i:],
                            int(math.floor(len(qry_rf_indel) / 3) * 3))

                    mut_name, aa_ref, aa_alt = PointFinder.name_indel_mutation(
                        sbjct_seq, indel, sbjct_rf_indel, qry_rf_indel,
                        codon_no, mut, sbjct_start - 1)

                    # Set index to the correct reading frame after the
                    # indel gap
                    shift_diff_before = abs(s_shift - q_shift)
                    s_shift += len(sbjct_rf_indel) - 3
                    q_shift += len(qry_rf_indel) - 3
                    shift_diff = abs(s_shift - q_shift)

                    if shift_diff_before != 0 and shift_diff % 3 == 0:

                        if s_shift > q_shift:
                            nucs_needed = (int((len(sbjct_rf_indel) / 3) * 3)
                                           + shift_diff)
                            pre_qry_indel = qry_rf_indel
                            qry_rf_indel = PointFinder.get_inframe_gap(
                                qry_seq[q_i:], nucs_needed)
                            q_shift += len(qry_rf_indel) - len(pre_qry_indel)

                        elif q_shift > s_shift:
                            nucs_needed = (int((len(qry_rf_indel) / 3) * 3)
                                           + shift_diff)
                            pre_sbjct_indel = sbjct_rf_indel
                            sbjct_rf_indel = PointFinder.get_inframe_gap(
                                sbjct_seq[s_i:], nucs_needed)
                            s_shift += (len(sbjct_rf_indel)
                                        - len(pre_sbjct_indel))

                        mut_name, aa_ref, aa_alt = (
                            PointFinder.name_indel_mutation(
                                sbjct_seq, indel, sbjct_rf_indel, qry_rf_indel,
                                codon_no, mut, sbjct_start - 1)
                        )
                        if "Frameshift" in mut_name:
                            mut_name = (mut_name.split("-")[0]
                                        + "- Frame restored")

                    mis_matches += [[mut, codon_no_indel, seq_pos, indel,
                                     mut_name, sbjct_rf_indel, qry_rf_indel,
                                     aa_ref, aa_alt]]

                    # Check if the next mutation in the indels list is
                    # in the current codon.
                    # Find the number of individul gaps in the
                    # evaluated sequence.
                    no_of_indels = (len(re.findall(r"\-\w", sbjct_rf_indel))
                                    + len(re.findall(r"\-\w", qry_rf_indel)))

                    if no_of_indels > 1:

                        for j in range(indel_no, indel_no + no_of_indels - 1):
                            try:
                                indel_data = indels[j]
                            except IndexError:
                                sys.exit("indel_data list is out of range, "
                                         "bug!")
                            mut = indel_data[0]
                            codon_no_indel = indel_data[1]
                            seq_pos = indel_data[2] + sbjct_start - 1
                            indel = indel_data[3]
                            indel_no += 1
                            mis_matches += [[mut, codon_no_indel, seq_pos,
                                             indel, mut_name, sbjct_rf_indel,
                                             qry_rf_indel, aa_ref, aa_alt]]

                    # Set codon number, and save nucleotides from out
                    # of frame mutations
                    if mut == "del":
                        codon_no += int((len(sbjct_rf_indel) - 3) / 3)
                    # If evaluated insert is only gaps codon_no should
                    # not increment
                    elif sbjct_rf_indel.count("-") == len(sbjct_rf_indel):
                        codon_no -= 1

                # Check of point mutations
                else:
                    mut = "sub"
                    aa_ref = PointFinder.aa(sbjct_codon)
                    aa_alt = PointFinder.aa(qry_codon)

                    if aa_ref != aa_alt:
                        # End search for mutation if a premature stop
                        # codon is found
                        mut_name = "p." + aa_ref + str(codon_no) + aa_alt

                        mis_matches += [[mut, codon_no, codon_no, aa_alt,
                                         mut_name, sbjct_codon, qry_codon,
                                         aa_ref, aa_alt]]
                # If a Premature stop codon occur report it an stop the
                # loop
                if not self.ignore_stop_codons:
                    try:
                        if mis_matches[-1][-1] == "*":
                            mut_name += " - Premature stop codon"
                            mis_matches[-1][4] = (
                                mis_matches[-1][4].split("-")[0]
                                + " - Premature stop codon")
                            break
                    except IndexError:
                        pass

        # Sort mutations on position
        mis_matches = sorted(mis_matches, key=lambda x: x[1])

        return mis_matches

    @staticmethod
    def mutstr2mutdict(m):
        out_dict = {}

        # Protein / Amino acid mutations
        # Ex.: "p.T83I"
        if(m.startswith("p.")):
            out_dict["nucleotide"] = False

            # Remove frameshift tag
            frameshift_match = re.search(r"(.+) - Frameshift.*$", m)
            if(frameshift_match):
                m = frameshift_match.group(1)
                out_dict["frameshift"] = True

            # Remove frame restored tag
            framerestored_match = re.search(r"(.+) - Frame restored.*$", m)
            if(framerestored_match):
                m = framerestored_match.group(1)
                out_dict["frame restored"] = True

            # Remove premature stop tag
            prem_stop_match = re.search(r"(.+) - Premature stop codon.*$", m)
            if(prem_stop_match):
                m = prem_stop_match.group(1)
                out_dict["prem_stop"] = True

            # TODO: premature or frameshift tag adds too many whitespaces
            m = m.strip()

            # Delins
            multi_delins_match = re.search(
                r"^p.(\D{1})(\d+)_(\D{1})(\d+)delins(\S+)$", m)
            single_delins_match = re.search(
                r"^p.(\D{1})(\d+)delins(\S+)$", m)
            # TODO: is both necessary?
            multi_delins_match2 = re.search(
                r"^p.(\D{1})(\d+)_(\D{1})(\d+)delins$", m)
            single_delins_match2 = re.search(
                r"^p.(\D{1})(\d+)delins$", m)
            multi_ins_match = re.search(
                r"^p.(\D{1})(\d+)_(\D{1})(\d+)ins(\D*)$", m)
            if(multi_delins_match or single_delins_match):
                out_dict["deletion"] = True
                out_dict["insertion"] = True
                if(single_delins_match):
                    out_dict["ref_aa"] = single_delins_match.group(1)
                    out_dict["pos"] = single_delins_match.group(2)
                    out_dict["mut_aa"] = single_delins_match.group(3)
                else:
                    out_dict["ref_aa"] = multi_delins_match.group(1)
                    out_dict["pos"] = multi_delins_match.group(2)
                    out_dict["ref_aa_right"] = multi_delins_match.group(3)
                    out_dict["mut_end"] = multi_delins_match.group(4)
                    out_dict["mut_aa"] = multi_delins_match.group(5)
            # Deletions
            elif(m[-3:] == "del"):
                single_del_match = re.search(
                    r"^p.(\D{1})(\d+)del$", m)
                multi_del_match = re.search(
                    r"^p.(\D{1})(\d+)_(\D{1})(\d+)del$", m)
                out_dict["deletion"] = True
                if(single_del_match):
                    out_dict["ref_aa"] = single_del_match.group(1)
                    out_dict["pos"] = single_del_match.group(2)
                else:
                    out_dict["ref_aa"] = multi_del_match.group(1)
                    out_dict["pos"] = multi_del_match.group(2)
                    out_dict["ref_aa_right"] = multi_del_match.group(3)
                    out_dict["mut_end"] = multi_del_match.group(4)
            # Insertions
            elif(multi_ins_match):
                out_dict["insertion"] = True
                out_dict["ref_aa"] = multi_ins_match.group(1).lower()
                out_dict["pos"] = multi_ins_match.group(2)
                out_dict["ref_aa_right"] = multi_ins_match.group(3).lower()
                if(multi_ins_match.group(4)):
                    out_dict["mut_aa"] = multi_ins_match.group(4).lower()
            # Substitutions
            else:
                sub_match = re.search(
                    r"^p.(\D{1})(\d+)(\D{1})$", m)
                out_dict["ref_aa"] = sub_match.group(1).lower()
                out_dict["pos"] = sub_match.group(2)
                out_dict["mut_aa"] = sub_match.group(3).lower()

        # Nucleotide mutations
        # Ex. sub: n.-42T>C
        # Ex. ins: n.-13_-14insG    (TODO: Verify)
        # Ex. del: n.42delT         (TODO: Verify)
        # Ex. del: n.42_45del       (TODO: Verify)
        elif(m.startswith("n.") or m.startswith("r.")):
            out_dict["nucleotide"] = True

            sub_match = re.search(
                r"^[nr]{1}\.(-{0,1}\d+)(\D{1})>(\D{1})$", m)
            ins_match = re.search(
                r"^[nr]{1}\.(-{0,1}\d+)_(-{0,1}\d+)ins(\S+)$", m)
            # r.541delA
            del_match = re.search((
                r"^[nr]{1}\.(-{0,1}\d+)_{0,1}(-{0,1}\d*)del(\S*)$"), m)
            if(sub_match):
                out_dict["pos"] = sub_match.group(1)
                out_dict["ref_nuc"] = sub_match.group(2)
                out_dict["mut_nuc"] = sub_match.group(3)
            elif(ins_match):
                out_dict["insertion"] = True
                out_dict["pos"] = ins_match.group(1)
                out_dict["mut_end"] = ins_match.group(2)
                out_dict["mut_nuc"] = ins_match.group(3)
            elif(del_match):
                out_dict["deletion"] = True
                out_dict["pos"] = del_match.group(1)
                if(del_match.group(2)):
                    out_dict["mut_end"] = del_match.group(2)
                if(del_match.group(3)):
                    out_dict["ref_nuc"] = del_match.group(3)
                else:
                    sys.exit("ERROR: Nucleotide deletion did not contain any "
                             "reference sequence. mut string: {}".format(m))

        return out_dict

    def get_mutations(self, gene, gene_name, mis_matches, unknown_flag, hit):
        RNA = False
        if gene in self.RNA_gene_list:
            RNA = True

        known_muts = []
        unknown_muts = []

        # Go through each mutation
        for i in range(len(mis_matches)):
            m_type = mis_matches[i][0]
            pos = mis_matches[i][1]  # sort on pos?
            look_up_pos = mis_matches[i][2]
            look_up_mut = mis_matches[i][3]
            mut_name = mis_matches[i][4]
            # nuc_ref = mis_matches[i][5]
            # nuc_alt = mis_matches[i][6]
            ref = mis_matches[i][-2]
            alt = mis_matches[i][-1]
            mut_dict = PointFinder.mutstr2mutdict(mut_name)

            mut_id = ("{gene}_{pos}_{alt}"
                      .format(gene=gene_name, pos=pos, alt=alt.lower()))

            ref_aa = mut_dict.get("ref_aa", None)
            ref_aa_right = mut_dict.get("ref_aa_right", None)
            mut_aa = mut_dict.get("mut_aa", None)
            ref_nuc = mut_dict.get("ref_nuc", None)
            mut_nuc = mut_dict.get("mut_nuc", None)
            is_nuc = mut_dict.get("nucleotide", None)
            is_ins = mut_dict.get("insertion", None)
            is_del = mut_dict.get("deletion", None)
            mut_end = mut_dict.get("mut_end", None)
            prem_stop = mut_dict.get("prem_stop", False)

            mut = ResMutation(unique_id=mut_id, seq_region=gene_name, pos=pos,
                              hit=hit, ref_codon=ref_nuc, mut_codon=mut_nuc,
                              ref_aa=ref_aa, ref_aa_right=ref_aa_right,
                              mut_aa=mut_aa, insertion=is_ins, deletion=is_del,
                              end=mut_end, nuc=is_nuc,
                              premature_stop=prem_stop, ab_class=None)

            if "Premature stop codon" in mut_name:
                sbjct_len = hit['sbjct_length']
                qry_len = pos * 3
                perc_trunc = round(
                    ((float(sbjct_len) - qry_len)
                     / float(sbjct_len))
                    * 100, 2
                )
                mut.premature_stop = perc_trunc

            # Check if mutation is known
            gene_mut_name, resistence, pmid = self.look_up_known_muts(
                gene, look_up_pos, look_up_mut, m_type, gene_name)

            # Collect known mutations
            if resistence != "Unknown":
                known_muts.append(mut)
            # Collect unknown mutations
            else:
                unknown_muts.append(mut)

            # TODO: Use ResMutation class to make sure identical mutations are
            #       not kept.

        return (known_muts, unknown_muts)

    def mut2str(self, gene, gene_name, mis_matches, unknown_flag, hit):
        """
            This function takes a gene name a list of mis matches found
            between subject and query of this gene, the dictionary of
            known mutation in the point finder database, and the flag
            telling weather the user wants unknown mutations to be
            reported. All mis matches are looked up in the known
            mutation dict to se if the mutation is known, and in this
            case what drug resistence it causes. The funtions return 3
            strings that are used as output to the users. One string is
            only tab seperated and contains the mutations listed line
            by line. If the unknown flag is set to true it will contain
            both known and unknown mutations. The next string contains
            only known mutation and are given in in a format that is
            easy to convert to HTML. The last string is the HTML tab
            sting from the unknown mutations.
        """
        known_header = ("Mutation\tNucleotide change\tAmino acid change\t"
                        "Resistance\tPMID\n")

        unknown_header = "Mutation\tNucleotide change\tAmino acid change\n"

        RNA = False
        if gene in self.RNA_gene_list:
            RNA = True
            known_header = "Mutation\tNucleotide change\tResistance\tPMID\n"
            unknown_header = "Mutation\tNucleotide change\n"

        known_lst = []
        unknown_lst = []
        all_results_lst = []
        output_mut = []
        stop_codons = []
        # Go through each mutation
        for i in range(len(mis_matches)):
            m_type = mis_matches[i][0]
            codon_pos = mis_matches[i][1]  # sort on codon_pos?
            nuc_pos = mis_matches[i][2]
            look_up_mut = mis_matches[i][3]
            mut_name = mis_matches[i][4]
            nuc_ref = mis_matches[i][5]
            nuc_alt = mis_matches[i][6]
            ref = mis_matches[i][-2]
            aa_alt = mis_matches[i][-1]

            # First index in list indicates if mutation is known
            output_mut += [[]]

            # Define output vaiables
            codon_change = nuc_ref + " -> " + nuc_alt
            aa_change = ref + " -> " + aa_alt

            if RNA is True:
                aa_change = "RNA mutations"
            elif codon_pos < 0:
                aa_change = "Promoter mutations"

            # Check if mutation is known
            gene_mut_name, resistence, pmid = self.look_up_known_muts(
                gene, nuc_pos, aa_alt, m_type, gene_name, nuc_alt, codon_pos,
                look_up_mut)

            # Make lists to strings
            if resistence != "Unknown":
                resistence = ", ".join(resistence).title()

            pmid = ", ".join(pmid)

            gene_mut_name = gene_mut_name + " " + mut_name

            output_mut[i] = [gene_mut_name, codon_change, aa_change,
                             resistence, pmid]

            # Add mutation to output strings for known mutations
            if resistence != "Unknown":
                if RNA is True:
                    # don't include the amino acid change field for
                    # RNA mutations
                    known_lst += ["\t".join(output_mut[i][:2]) + "\t"
                                  + "\t".join(output_mut[i][3:])]
                else:
                    known_lst += ["\t".join(output_mut[i])]

                all_results_lst += ["\t".join(output_mut[i])]

            # Add mutation to output strings for unknown mutations
            else:
                if RNA is True:
                    unknown_lst += ["\t".join(output_mut[i][:2])]
                else:
                    unknown_lst += ["\t".join(output_mut[i][:3])]

                if unknown_flag is True:
                    all_results_lst += ["\t".join(output_mut[i])]


            # Check that you do not print two equal lines (can happen
            # if two indels occure in the same codon)
            if len(output_mut) > 1:
                if output_mut[i] == output_mut[i - 1]:

                    if resistence != "Unknown":
                        known_lst = known_lst[:-1]
                        all_results_lst = all_results_lst[:-1]
                    else:
                        unknown_lst = unknown_lst[:-1]

                        if unknown_flag is True:
                            all_results_lst = all_results_lst[:-1]

            if "Premature stop codon" in mut_name:
                sbjct_len = hit['sbjct_length']
                qry_len = codon_pos * 3
                prec_truckat = round(
                    ((float(sbjct_len) - qry_len)
                     / float(sbjct_len))
                    * 100, 2
                )
                perc = "%"
                stop_codons.append("Premature stop codon in %s, %.2f%s lost"
                                   % (gene, prec_truckat, perc))

        # Creat final strings
        if(all_results_lst):
            all_results = "\n".join(all_results_lst)
        else:
            all_results = ""
        total_known_str = ""
        total_unknown_str = ""

        # Check if there are only unknown mutations
        resistence_lst = []
        for mut in output_mut:
            for res in mut[3].split(","):
                res = res.lstrip()
                resistence_lst.append(res)

        # Save known mutations
        unknown_no = resistence_lst.count("Unknown")
        if unknown_no < len(resistence_lst):
            total_known_str = known_header + "\n".join(known_lst)
        else:
            total_known_str = "No known mutations found in %s" % gene_name

        # Save unknown mutations
        if unknown_no > 0:
            total_unknown_str = unknown_header + "\n".join(unknown_lst)
        else:
            total_unknown_str = "No unknown mutations found in %s" % gene_name

        return (all_results, total_known_str, total_unknown_str,
                resistence_lst + stop_codons)

    @staticmethod
    def get_file_content(file_path, fst_char_only=False):
        """
        This function opens a file, given as the first argument
        file_path and returns the lines of the file in a list or the
        first character of the file if fst_char_only is set to True.
        """
        with open(file_path, "r") as infile:
            line_lst = []
            for line in infile:
                line = line.strip()
                if line != "":
                    line_lst.append(line)
                if fst_char_only:
                    return line_lst[0][0]
        return line_lst

    def look_up_known_muts(self, gene, nuc_pos, found_mut, mut, gene_name,
                           found_nuc, codon_pos, found_nuc_del):
        """
            This function uses the known_mutations dictionay, a gene a
            string with the gene key name, a gene position as integer,
            found_mut is given as amino acid or nucleotides, the
            mutation type (mut) is either "del", "ins", or "sub", and
            gene_name is the gene name that should be returned to the
            user. The function looks up if the found_mut defined by the
            gene, position and the found_mut string is given in the
            known_mutations dictionary, if it is, the resistance and
            the pmid are returned together with the gene_name given in
            the known_mutations dict. If the mutation type is "del" the
            deleted nucleotids are checked to be contained in any of
            the deletion described in the known_mutation dict.
        """
        resistence = "Unknown"
        pmid = "-"
        gene_ID = gene.split("_")[0]
        found_nuc = found_nuc.upper()
        found_mut = found_mut.upper()

        if mut == "del":
            found_mut = found_nuc_del.upper()
            for i, i_pos in enumerate(range(nuc_pos, nuc_pos + len(found_mut))):

                known_indels = self.known_mutations[gene_ID]["del"].get(i_pos,
                                                                        [])
                for known_indel in known_indels:
                    partial_mut = found_mut[i:len(known_indel) + i]

                    # Check if part of found mut is known and check if
                    # found mut and known mut is in the same reading
                    # frame
                    if (partial_mut == known_indel
                            and len(found_mut) % 3 == len(known_indel) % 3):
                        resistence = (
                        self.known_mutations[gene_ID]["del"][i_pos]
                        [known_indel]['drug'])

                        pmid = (self.known_mutations[gene_ID]["del"][i_pos]
                        [known_indel]['pmid'])

                        gene_name = (self.known_mutations[gene_ID]["del"][i_pos]
                        [known_indel]['gene_name'])
                        break
        elif mut == 'sub':
            if codon_pos in self.known_mutations[gene_ID][mut]:
                if found_mut in self.known_mutations[gene_ID][mut][codon_pos]:
                    resistence = (self.known_mutations[gene_ID][mut][codon_pos]
                    [found_mut]['drug'])

                    pmid = (self.known_mutations[gene_ID][mut][codon_pos]
                    [found_mut]['pmid'])

                    gene_name = (self.known_mutations[gene_ID][mut][codon_pos]
                    [found_mut]['gene_name'])
        else:
            if (nuc_pos in self.known_mutations[gene_ID][mut]
                    or codon_pos in self.known_mutations[gene_ID][mut]):
                found_mutation = None
                pos = None
                if (codon_pos in self.known_mutations[gene_ID][mut]
                        and found_mut in self.known_mutations[gene_ID][mut][
                            codon_pos]):
                    found_mutation = found_mut
                    pos = codon_pos
                if (nuc_pos in self.known_mutations[gene_ID][mut]
                        and found_nuc in self.known_mutations[gene_ID][mut][
                            nuc_pos]):
                    found_mutation = found_nuc
                    pos = nuc_pos
                if found_mutation:
                    resistence = (self.known_mutations[gene_ID][mut][pos]
                                  [found_mutation]['drug'])

                    pmid = (self.known_mutations[gene_ID][mut][pos]
                            [found_mutation]['pmid'])

                    gene_name = (self.known_mutations[gene_ID][mut][pos]
                                [found_mutation]['gene_name'])

        # Check if stop codons refer resistance
        if "*" in found_mut and gene_ID in self.known_stop_codon:
            if resistence == "Unknown":
                resistence = self.known_stop_codon[gene_ID]["drug"]
            else:
                resistence += "," + self.known_stop_codon[gene_ID]["drug"]

        return (gene_name, resistence, pmid)


if __name__ == '__main__':

    ##########################################################################
    # PARSE COMMAND LINE OPTIONS
    ##########################################################################

    parser = argparse.ArgumentParser(
        description="This program predicting resistance associated with \
                     chromosomal mutations based on WGS data",
        prog="pointfinder.py")

    # required arguments
    parser.add_argument("-i",
                        dest="inputfiles",
                        metavar="INFILE",
                        nargs='+',
                        help="Input file. fastq file(s) from one sample for \
                              KMA or one fasta file for blastn.",
                        required=True)
    parser.add_argument("-o",
                        dest="out_path",
                        metavar="OUTFOLDER",
                        help="Output folder, output files will be stored \
                              here.",
                        required=True)
    parser.add_argument("-s",
                        dest="species",
                        metavar="SPECIES",
                        choices=['e.coli', 'gonorrhoeae', 'campylobacter',
                                 'salmonella', 'tuberculosis'],
                        help="Species of choice, e.coli, tuberculosis, \
                              salmonella, campylobactor, gonorrhoeae, \
                              klebsiella, or malaria",
                        required=True)
    parser.add_argument("-m",
                        dest="method",
                        metavar="METHOD",
                        choices=["kma", "blastn"],
                        help="Method of choice, kma or blastn",
                        required=True)
    parser.add_argument("-m_p",
                        dest="method_path",
                        help="Path to the location of blastn or kma dependent \
                              of the chosen method",
                        required=True)
    parser.add_argument("-p",
                        dest="db_path",
                        metavar="DATABASE",
                        help="Path to the location of the pointfinder \
                              database",
                        required=True)

    # optional arguments
    parser.add_argument("-t",
                        dest="threshold",
                        metavar="IDENTITY",
                        help="Minimum gene identity threshold, default = 0.9",
                        type=float,
                        default=0.9)
    parser.add_argument("-l",
                        dest="min_cov",
                        metavar="COVERAGE",
                        help="Minimum gene coverage threshold, \
                              threshold = 0.6",
                        type=float,
                        default=0.6)
    parser.add_argument("-u",
                        dest="unknown_mutations",
                        help="Show all mutations found even if it's unknown \
                              to the resistance database.",
                        action='store_true',
                        default=False)
    parser.add_argument("-g",
                        dest="specific_gene",
                        nargs='+',
                        help="Specify genes existing in the database to \
                              search for - if none is specified all genes are \
                              included in the search.",
                        default=None)
    parser.add_argument("-ii",
                        dest="ignore_indels",
                        metavar="IGNORE_INDELS",
                        action="store_true",
                        help="Assume all insertions and deletions are \
                              sequencing errors. Don't account for \
                              frameshifts caused by indels.",
                        default=False)
    parser.add_argument("-ic",
                        dest="ignore_stop_codons",
                        metavar="IGNORE_STOP_CODONS",
                        action="store_true",
                        help="Assume premature stop codons are a result of \
                              sequencing error. Detect mutations that \
                              occur after premature stop codon.",
                        default=False)

    args = parser.parse_args()

    # If no arguments are given print usage message and exit
    if len(sys.argv) == 1:
        sys.exit("Usage: " + parser.usage)

    if(args.method == "blastn"):
        method = PointFinder.TYPE_BLAST
    else:
        method = PointFinder.TYPE_KMA

    # Get sample name
    filename = args.inputfiles[0].split("/")[-1]
    sample_name = "".join(filename.split(".")[0:-1])  # .split("_")[0]
    if sample_name == "":
        sample_name = filename

    # TODO: Change ilogocal var name
    kma_db_path = args.db_path + "/" + args.species

    finder = PointFinder(db_path=kma_db_path, species=args.species,
                         gene_list=args.specific_gene,
                         ignore_indels=args.ignore_indels,
                         ignore_stop_codons=args.ignore_stop_codons)

    if method == PointFinder.TYPE_BLAST:

        # Check that only one input file is given
        if len(args.inputfiles) != 1:
            sys.exit("Input Error: Blast was chosen as mapping method only 1 "
                     "input file requied, not %s" % (len(args.inputfiles)))

        finder_run = finder.blast(inputfile=args.inputfiles[0],
                                 out_path=args.out_path,
                                 min_cov=0.01,
                                 threshold=args.threshold,
                                 blast=args.method_path,
                                 cut_off=False)
    else:
        inputfile_1 = args.inputfiles[0]
        inputfile_2 = None

        if(len(args.inputfiles) == 2):
            inputfile_2 = args.inputfiles[1]
        finder_run = finder.kma(inputfile_1=inputfile_1,
                             inputfile_2=inputfile_2,
                             out_path=args.out_path,
                             db_path_kma=kma_db_path,
                             databases=[args.species],
                             min_cov=args.min_cov,
                             threshold=args.threshold,
                             kma_path=args.method_path,
                             sample_name=sample_name,
                             kma_mrs=0.5, kma_gapopen=-5, kma_gapextend=-2,
                             kma_penalty=-3, kma_reward=1)

    results = finder_run.results

    if(args.specific_gene):
        results = PointFinder.discard_unwanted_results(
            results=results, wanted=args.specific_gene)

    finder.write_results(out_path=args.out_path, result=results, min_cov=args.min_cov,
                         res_type=method, unknown_flag=args.unknown_mutations)



#!/usr/bin/env python3
import random
import string
import json
import re

from ..phenotype2genotype.res_profile import PhenoDB
from ..phenotype2genotype.feature import ResGene, ResMutation

from .exceptions import DuplicateKeyError
from .gene_result import GeneResult
from .seq_variation_result import SeqVariationResult
from .phenotype_result import PhenotypeResult

from ..pointfinder import PointFinder

def add_gene_result_if_key_not_None(gene_result, res_collection):
    '''
        Input:
            gene_result: GeneResult object (seq_region dict)
            res_collection: Result object created by the cge core module.
        adds gene_result to res_collection if gene_result['key'] is different
        from None.
            '''
    if gene_result["key"] is None:
        return
    elif gene_result["key"] not in res_collection["seq_regions"]:
        res_collection.add_class(cl="seq_regions", **gene_result)
    else:
        raise DuplicateKeyError(
            "About to overwrite dict entry. This should not be "
            "happening as all keys are supposed to be unique."
            "Non-unique key was: {}".format(gene_result["key"]))



class ResFinderResultHandler():

    @staticmethod
    def standardize_results(res_collection, res, ref_db_name, conf):
        """
            Input:
                res_collection: Result object created by the cge core module.
                res: Custom dictionary of results from ResFinder
                ref_db_name: 'ResFinder' or 'PointFinder' or 'DisinFinder'

            Method loads the given res_collection with results from res.
        """
        for db_name, db in res.results.items():
            if(db_name == "excluded"):
                continue

            if(db == "No hit found"):
                continue

            for unique_id, hit_db in db.items():
                if(unique_id in res.results["excluded"]):
                    continue
                gene_result = GeneResult(res_collection, hit_db, ref_db_name,
                                         conf, res)

                add_gene_result_if_key_not_None(gene_result, res_collection)

    @staticmethod
    def load_res_profile(res_collection, isolate, amr_abbreviations):
        """
            Input:
                res_collection: Result object created by the cge core module.
                isolate: Isolate object

            Method loads the given res_collection with results from res.
        """
        # For each antibiotic class
        for ab_class in isolate.resprofile.phenodb.antibiotics.keys():
            # For each antibiotic in current class
            for phenodb_ab in isolate.resprofile.phenodb.antibiotics[ab_class]:
                phenotype = PhenotypeResult(phenodb_ab, isolate)

                # Isolate is resistant towards the antibiotic
                if(phenodb_ab in isolate.resprofile.resistance):
                    ResFinderResultHandler._load_resistant_phenotype(
                        phenotype, isolate, phenodb_ab, res_collection)

                res_collection.add_class(cl="phenotypes", **phenotype)

        amr_sum = ResFinderResultHandler.create_amr_summary_str(
            res_collection, amr_abbreviations)
        res_collection.add(**{"result_summary": amr_sum})

    @staticmethod
    def _load_resistant_phenotype(phenotype, isolate, antibiotic,
                                  res_collection):
        phenotype.set_resistant(True)
        isolate_ab = isolate.resprofile.resistance[antibiotic]

        for feature_lst in isolate_ab.features.values():
            for feature_entry in feature_lst:
                # feature_entry is either a Feature object or a dict of Feature
                # objects.
                try:
                    for unique_id, feature in feature_entry.items():
                        ResFinderResultHandler._add_feat_to_phenotype_and_res(
                            res_collection, isolate, feature, phenotype)
                # Feature object has no items attribute.
                except AttributeError:
                    ResFinderResultHandler._add_feat_to_phenotype_and_res(
                        res_collection, isolate, feature_entry, phenotype)

    @staticmethod
    def _add_feat_to_phenotype_and_res(res_collection, isolate, feature,
                                       phenotype):
        if isinstance(feature, ResGene) or isinstance(feature, ResMutation):
            if isinstance(feature, ResMutation):
                if feature.unique_id in phenotype.get("seq_variations", ()):
                    return
            phenotype.add_feature(res_collection, isolate, feature)

    @staticmethod
    def create_amr_summary_str(res_collection, amr_abbreviations):
        amr_list = []
        for key, phenotype in res_collection["phenotypes"].items():
            if(phenotype["amr_resistant"] is True
               and phenotype["amr_species_relevant"] is True):
                amr_name = phenotype["amr_resistance"].capitalize()
                amr = amr_abbreviations.get(amr_name,
                                            [phenotype["amr_resistance"]])
                amr_list.append(amr[0])
        if(amr_list):
            out_str = "_".join(amr_list)
        else:
            out_str = ""
        return out_str


class PointFinderResultHandler():

    @staticmethod
    def standardize_results(res_collection, res, ref_db_name):
        """
            Input:
                res_collection: Result object created by the cge core module.
                res: Custom dictionary of results from PointFinder
                ref_db_name: 'ResFinder' or 'PointFinder' or 'DisinFinder'

            Method loads the given res_collection with results from res.
        """
        for gene_name, db in res.items():
            # Ignore information in excluded dict
            if(gene_name == "excluded"):
                continue

            # Ignore genes found in excluded dict
            if gene_name in res["excluded"]:
                continue
            if(isinstance(db, str)):
                if(db == "No hit found"):
                    continue
                if db.startswith("Gene found with coverage"):
                    continue

            gene_results = []

            # For BLAST results
            db_hits = db.get("hits", {})

            # For KMA results
            if(not db_hits):
                id = db["sbjct_header"]
                db_hits[id] = db

            for unique_id, hit_db in db_hits.items():

                # Ignore genes found in excluded dict
                if(unique_id in res["excluded"]):
                    continue
                gene_result = GeneResult(res_collection, hit_db, ref_db_name)

                add_gene_result_if_key_not_None(gene_result, res_collection)
                if gene_result['key'] is None:
                    continue

                gene_results.append(gene_result)

                # KMA hits
                mismatches = hit_db.get("mis_matches", None)
                # BLAST hits
                if(mismatches is None):
                    mismatches = db["mis_matches"]

                for mismatch in mismatches:
                    seq_var_result = SeqVariationResult(
                        res_collection, mismatch, gene_results, ref_db_name)
                    res_collection.add_class(cl="seq_variations",
                                                **seq_var_result)

    @staticmethod
    def rename_promoter_results(results, method):
        if method == PointFinder.TYPE_BLAST:
            PointFinderResultHandler._rename_blast_promoter_results(results)
        else:
            PointFinderResultHandler._rename_kma_promoter_results(results)
    
    def _rename_blast_promoter_results(results):
        promoter_keys = [k for k in results if re.search(r"promoter-size", k)]
        for promoter_key in promoter_keys:
            val_dict = results.pop(promoter_key)
            new_promoter_key = PhenoDB.if_promoter_gene_rename(promoter_key)
            results[new_promoter_key] = val_dict
    
    def _rename_kma_promoter_results(results):
        promoter_keys = set()

        for inner_dict in results.values():
            promoter_keys = [k for k in inner_dict if re.search(r"promoter-size", k)]
            for promoter_key in promoter_keys:
                val_dict = inner_dict.pop(promoter_key)
                new_promoter_key = PhenoDB.if_promoter_gene_rename(promoter_key)
                inner_dict[new_promoter_key] = val_dict



#!/usr/bin/env python3

import sys
import re
from string import Template


class PanelNameError(Exception):
    """ Raise when panel name given does not match panel names from file.
    """
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(PanelNameError, self).__init__(message, *args)


class ResSumTable(dict):
    """
    """
    def __init__(self, text):
        """
        """
        self.panels = {}
        self.inclusions = {}
        self.name = ""

        has_missing_features = False
        self.missing_features = []

        for line in text.splitlines():

            # Skip Comments
            if(line.startswith("#")):
                # Storing missing feature headers
                if(has_missing_features):
                    self.missing_features.append(line)

                if(line.startswith("# Sample: ")):
                    self.name = line[10:].strip()
                # Only exists if the table has a missing feature warning
                elif(line.startswith("# WARNING")):
                    has_missing_features = True
                    self.missing_features.append(line)
                continue

            # Skip empty lines
            if(not line):
                continue

            if(has_missing_features):
                self.missing_features.append(line)
            else:
                line_list = line.split("\t")
                self[line_list[0]] = line_list

    def _merge_inclusions(self):
        """
        """
        for panel in self.inclusions:
            panel_list = self.panels[panel]
            include_list = self.inclusions[panel]
            include_abs = []
            for incl_panel in include_list:
                include_abs = include_abs + self.panels[incl_panel]
            self.panels[panel] = panel_list + include_abs

    def _remove_redundancy(self):
        for panel in self.panels:
            panel_list = self.panels[panel]
            self.panels[panel] = list(set(panel_list))

    def load_amr_panels(self, panel_file):
        """
        """
        with open(panel_file, "r") as fh:
            panel_name = None
            re_panel = re.compile(r':Panel:\s{0,1}(.+)$')
            re_include = re.compile(r':Include:\s{0,1}(.+)$')
            for line in fh:

                line = line.rstrip()

                # Skip empty lines
                if(not line):
                    continue

                # Skip Comments
                if(line.startswith("#")):
                    continue

                # Get panel name
                match_panel = re_panel.search(line)
                if(match_panel):
                    panel_name = match_panel.group(1).lower()
                    self.panels[panel_name] = []
                    species_panel_name = None
                    continue

                # Get inclusions
                match_inclusion = re_include.search(line)
                if(match_inclusion):
                    include_panel = match_inclusion.group(1).lower()
                    tmp_list = []

                    if(panel_name in self.inclusions):
                        tmp_list = self.inclusions[panel_name]

                    tmp_list.append(include_panel)
                    self.inclusions[panel_name] = tmp_list

                    continue

                # Antibiotics
                # Stores list of antimicrobials in dict of panel names.
                if(panel_name):
                    if(panel_name in self.panels):
                        tmp_list = self.panels[panel_name]
                        tmp_list.append(line.lower())
                        self.panels[panel_name] = tmp_list
                    else:
                        tmp_list = [line.lower()]
                        self.panels[panel_name] = tmp_list

        self._merge_inclusions()
        self._remove_redundancy()

    def check_panel_name(self, name):
        """ Panel names are expected to consist of "Genus species" or
            only "Genus". The method checks the name against the loaded
            panel names and returns the panel name that matches the
            given name. If no panel name matches it checks if the first
            word of the given name matches any of the genus panel
            names, and considers that a match if found.

            Returns False if no match is found
            Returns None if no panels has been loaded
            Returns the panel name that match
        """
        if(not self.panels):
            return None

        name = name.lower()
        if(name in self.panels):
            return name

        if(name == "complete"):
            return name

        genus_name = " ".split(name)[0]
        if(genus_name in self.panels):
            return genus_name

        return False

    def get_amr_panel_str(self, panel_name_raw, header=False):
        """
        """
        panel_name = self.check_panel_name(name=panel_name_raw)
        if(not panel_name):
            raise(PanelNameError("ERROR: Panel name given was not found among "
                                 "the loaded panels. Panel name given: "
                                 + str(panel_name_raw)))

        output_str = ""

        if(header):
            output_str = (
                "# ResFinder phenotype results for " + panel_name + ".\n"
                "# Sample: " + self.name + "\n"
                "# \n"
                "# The phenotype 'No resistance' should be interpreted with\n"
                "# caution, as it only means that nothing in the used\n"
                "# database indicate resistance, but resistance could exist\n"
                "# from 'unknown' or not yet implemented sources.\n"
                "# \n"
                "# The 'Match' column stores one of the integers 0, 1, 2, 3.\n"
                "#      0: No match found\n"
                "#      1: Match < 100% ID AND match length < ref length\n"
                "#      2: Match < 100% ID AND match length = ref length\n"
                "#      3: Match = 100% ID AND match length = ref length\n"
                "# If several hits causing the same resistance are found,\n"
                "# the highest number will be stored in the 'Match' column.\n"
                "\n"
            )
            output_str += ("# Antimicrobial\t"
                           "Class\t"
                           "WGS-predicted phenotype\t"
                           "Match\t"
                           "Genetic background\n")

        for ab in self.panels[panel_name]:
            na_list = [ab, "NA", "NA", "NA", "Not in database"]
            ab_list = self.get(ab, na_list)
            output_str += "\t".join(ab_list) + "\n"

        if(self.missing_features):
            output_str += "\n" + "\n".join(self.missing_features)

        return output_str

    def get_html_panel_table(self, panel_name_raw, panel_id, indent='      '):
        panel_name = self.check_panel_name(name=panel_name_raw)
        if(not panel_name):
            raise(PanelNameError("ERROR: Panel name given was not found among "
                                 "the loaded panels. Panel name given: "
                                 + str(panel_name_raw)))

        beg = Template(
            '$indent<div id="$panel_name" class="tabcontent">\n'
            '$indent  <table id="$panel_id" rules="cols">\n'
            '$indent    <thead>\n'
            '$indent      <tr>\n'
            '$indent        <th class="head-ab-col" onclick="sortTable(0, '
            '\'$panel_id\')">Antimicrobial</th>\n'
            '$indent        <th class="head-col" onclick="sortTable(1, '
            '\'$panel_id\')">Class</th>\n'
            '$indent        <th class="head-col" onclick="sortTable(2, '
            '\'$panel_id\')">WGS-predicted phenotype</th>\n'
            '$indent        <th class="head-gene-col" onclick="sortTable(3, '
            '\'$panel_id\')">Genetic background</th>\n'
            '$indent      </tr>\n'
            '$indent    </thead>\n\n')

        output_str = beg.substitute(panel_name=panel_name,
                                    panel_id=panel_id,
                                    indent=indent)

        ab_row = Template(
            '$indent      <tr>\n'
            '$indent        <td class="ab-col">$ab</td>\n'
            '$indent        <td class="class-col">$abclass</td>\n'
            '$indent        <td class="$res">$pheno</td>\n'
            '$indent        <td class="genes-col">$genes</td>\n'
            '$indent      </tr>\n')

        output_str += indent + '    <tbody>\n'

        if(panel_name == "complete"):
            for ab in self:
                na_list = [ab, "NA", "NA", "NA", "Not in database"]
                ab_list = self.get(ab, na_list)
                if(ab_list[3] == "1"):
                    res = 'bggrey'
                elif(ab_list[3] == "2"):
                    res = 'bglightgreen'
                elif(ab_list[3] == "3"):
                    res = 'bggreen'
                else:
                    res = 'sus-col'
                output_str += ab_row.substitute(indent=indent,
                                                ab=ab,
                                                abclass=ab_list[1],
                                                res=res,
                                                pheno=ab_list[2],
                                                genes=ab_list[4])
        else:
            for ab in self.panels[panel_name]:
                na_list = [ab, "NA", "NA", "NA", "Not in database"]
                ab_list = self.get(ab, na_list)
                if(ab_list[3] == "1"):
                    res = 'bggrey'
                elif(ab_list[3] == "2"):
                    res = 'bglightgreen'
                elif(ab_list[3] == "3"):
                    res = 'bggreen'
                else:
                    res = 'sus-col'
                output_str += ab_row.substitute(indent=indent,
                                                ab=ab,
                                                abclass=ab_list[1],
                                                res=res,
                                                pheno=ab_list[2],
                                                genes=ab_list[4])

        output_str += (indent + '    </tbody>\n\n'
                       + indent + '  </table>\n'
                       + indent + '</div>\n')

        return output_str
